#Copyright 2021 Google LLC

#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
############################################################################################
########################################################################################################################
###### Many thanks to Hasan Rafiq for his excellent tutorial on Tensorflow pipelines where I learnt many helpful hints:
###### https://colab.research.google.com/gist/rafiqhasan/6f00aecf1feafd83ba9dfefef8907ee8/dl-e2e-taxi-dataset-tf2-keras.ipynb
###### Watch the entire video below on Srivatsan Srinivasan's excellent YouTube channel: AI Engineering   ##############
######                       https://youtu.be/wPri78CFSEw                                                 ##############
########################################################################################################################
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import tempfile
import pdb
import copy
import warnings
warnings.filterwarnings(action='ignore')
import functools
# Make numpy values easier to read.
np.set_printoptions(precision=3, suppress=True)
############################################################################################
# data pipelines and feature engg here

# pre-defined TF2 Keras models and your own models here

# Utils

############################################################################################
# TensorFlow ≥2.4 is required
import tensorflow as tf
np.random.seed(42)
tf.random.set_seed(42)
from tensorflow.keras import layers
from tensorflow import keras
from tensorflow.keras.layers.experimental.preprocessing import Normalization, StringLookup, Hashing
from tensorflow.keras.layers.experimental.preprocessing import IntegerLookup, CategoryEncoding, CategoryCrossing
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization, Discretization
from tensorflow.keras.layers import Embedding, Flatten

from tensorflow.keras.optimizers import SGD, Adam, RMSprop
from tensorflow.keras import layers
from tensorflow.keras import optimizers
from tensorflow.keras.models import Model, load_model
from tensorflow.keras import callbacks
from tensorflow.keras import backend as K
from tensorflow.keras import utils
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.optimizers import SGD
from tensorflow.keras import regularizers

from sklearn.metrics import roc_auc_score, mean_squared_error, mean_absolute_error
from IPython.core.display import Image, display
import pickle
#############################################################################################
##### Suppress all TF2 and TF1.x warnings ###################
tf2logger = tf.get_logger()
tf2logger.warning('Silencing TF2.x warnings')
tf2logger.root.removeHandler(tf2logger.root.handlers)
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
########################################################################################################################
from tensorflow.keras.layers import Reshape, MaxPooling1D, MaxPooling2D, AveragePooling2D, AveragePooling1D
from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Activation, Dense, Embedding, GlobalAveragePooling1D, GlobalMaxPooling1D, Dropout, Conv1D
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization
#### probably the most handy function of all!
def left_subtract(l1,l2):
    lst = []
    for i in l1:
        if i not in l2:
            lst.append(i)
    return lst
#############################################################################################################################
###### Many thanks to ML Design Patterns by Lak Lakshmanan which provided the following date-time TF functions below:
######    You can find more at : https://github.com/GoogleCloudPlatform/ml-design-patterns/tree/master/02_data_representation
############################################################################################################################
import datetime
def get_dayofweek(s):
    DAYS = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat','Sun']
    ts = parse_datetime(s)
    return DAYS[ts.weekday()]

def get_monthofyear(s):
    MONTHS = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun','Jul','Aug','Sep','Oct','Nov','Dec']
    ts = parse_datetime(s)
    return MONTHS[ts.month-1]

def get_hourofday(s):
    ts = parse_datetime(s)
    return str(ts.hour)

@tf.function
def dayofweek(ts_in):
    """
    This function converts dayofweek as a number to a string such as 4 means Thursday in dayofweek format.
    """
    return tf.map_fn(
        lambda dayofweek_number: tf.py_function(get_dayofweek, inp=[dayofweek_number], Tout=tf.string),
        ts_in)

@tf.function
def hourofday(ts_in):
    """
    This function converts dayofweek as a number to a string such as 4 means Thursday in dayofweek format.
    """
    return tf.map_fn(
        lambda dayofweek_number: tf.py_function(get_hourofday, inp=[dayofweek_number], Tout=tf.string),
        ts_in)

@tf.function
def monthofyear(ts_in):
    """
    This function converts dayofweek as a number to a string such as 4 means Thursday in dayofweek format.
    """
    return tf.map_fn(
        lambda dayofweek_number: tf.py_function(get_monthofyear, inp=[dayofweek_number], Tout=tf.string),
        ts_in)


def parse_datetime(timestring):
    if type(timestring) is not str:
        timestring = timestring.numpy().decode('utf-8') # if it is a Tensor
    return pd.to_datetime(timestring, infer_datetime_format=True, errors='coerce')

##########################################################################################################
from itertools import combinations
from collections import defaultdict
import copy
import time
def preprocessing_tabular(train_ds, var_df, cat_feature_cross_flag, model_options, cat_vocab_dict,
                                keras_model_type,verbose=0):
    """
    ############################################################################################
    # This preprocessing layer returns a tuple (all_features, all_inputs) as arguments to create_model function
    # You must then create a Functional model by transforming all_features into outputs like this:
    # The final step in create_model will use all_inputs as inputs
        x = tf.keras.layers.Dense(32, activation="relu")(all_features)
        x = tf.keras.layers.Dropout(0.5)(x)
        output = tf.keras.layers.Dense(1)(x)
        model = tf.keras.Model(all_inputs, output)
    ############################################################################################
    """
    start_time = time.time()
    drop_cols = var_df['cols_delete']
    train_data_is_file = model_options['train_data_is_file']
    #########  Now that you have the variable classification dictionary, just separate them out! ##
    cats = var_df['categorical_vars']  ### these are low cardinality vars - you can one-hot encode them ##
    high_string_vars = var_df['discrete_string_vars']  ## discrete_string_vars are high cardinality vars ## embed them!
    bools = var_df['bools']
    int_bools = var_df['int_bools']
    int_cats = var_df['int_cats'] + var_df['int_bools']
    ints = var_df['int_vars']
    floats = var_df['continuous_vars']
    nlps = var_df['nlp_vars']
    idcols = var_df['IDcols']
    dates = var_df['date_vars']
    lats = var_df['lat_vars']
    lons = var_df['lon_vars']
    matched_lat_lons = var_df['matched_pairs']
    floats = left_subtract(floats, lats+lons)

    #### These are the most important variables from this program: all inputs and outputs
    all_inputs = []
    all_encoded = []
    all_features = []
    all_input_names = []
    dropout_rate = 0.1 ### this is needed for feature crossing

    ### just use this to set the limit for max tokens for different variables ###
    ### we are setting the number of max_tokens to be 2X the number of tokens found in train
    max_tokens_zip = defaultdict(int)
    ##### Just combine Boolean and cat variables here to set the vocab ####
    cats_copy = copy.deepcopy(cats)
    if len(cats_copy) > 0:
        for each_name in cats_copy:
            max_tokens_zip[each_name] = cat_vocab_dict[each_name]['vocab'] ### just send vocab in
    high_cats_copy = copy.deepcopy(high_string_vars)
    if len(high_cats_copy) > 0:
        for each_name in high_cats_copy:
            max_tokens_zip[each_name] = cat_vocab_dict[each_name]['vocab'] ### just send vocab in
    copy_int_cats = copy.deepcopy(int_cats)
    if len(copy_int_cats) > 0:
        for each_int in copy_int_cats:
            max_tokens_zip[each_int] = cat_vocab_dict[each_int]['vocab'].tolist() ### just send vocab in
    copy_int_bools = copy.deepcopy(int_bools)
    if len(copy_int_bools) > 0:
        for each_int in copy_int_bools:
            max_tokens_zip[each_int] = cat_vocab_dict[each_int]['vocab'].tolist() ### just send vocab in
    copy_floats = copy.deepcopy(floats+bools)
    if len(copy_floats) > 0:
        for each_float in copy_floats:
            max_tokens_zip[each_float] = cat_vocab_dict[each_float]['vocab_min_var'] ### just send vocab as its a list
    copy_ints = copy.deepcopy(ints)
    if len(copy_ints) > 0:
        for each_int in copy_ints:
            max_tokens_zip[each_int] = int(1*(cat_vocab_dict[each_int]['size_of_vocab'])) ### size of vocab here
    if verbose > 1:
        print('    Number of categories in:')
        for each_name in max_tokens_zip.keys():
            if not each_name in high_cats_copy:
                print('        %s: %s' %(each_name, max_tokens_zip[each_name]))
            else:
                continue
    
    ####### CAVEAT : All the inputs and outputs should follow this same sequence below! ######
    all_date_inputs = []
    all_bool_inputs = []
    all_int_inputs = []
    all_int_cat_inputs = []
    all_cat_inputs = []
    all_num_inputs = []
    all_latlon_inputs = []
    ############## CAVEAT: This is the new way of concatenating different kinds of variables together ####
    all_new_cat_encoded = [] ## this is the new way of encoding cat vars and tying them together ###
    all_new_numeric_encoded = [] ### This is the new way of encoding num vars and tying them together ##

    ############## CAVEAT: The encoded outputs should follow the same sequence as inputs above!
    all_date_encoded = []
    all_bool_encoded = []
    all_int_bool_encoded = []
    all_int_encoded = []
    all_int_cat_encoded = []
    all_cat_encoded = []
    all_high_cat_encoded = []
    all_feat_cross_encoded = []
    all_num_encoded = []
    all_latlon_encoded = []
    lat_lon_paired_encoded = []
    cat_encoded_dict = dict([])
    cat_input_dict = dict([])
    date_input_dict = dict([])
    bool_input_dict = dict([])
    ###############################
    high_cats_alert = 50 ### set this number to alery you when a variable has high dimensions. Should it?
    hidden_units = [32, 32]  ## this is the number needed for feature crossing
    ####### We start creating variables encoding with date-time variables first ###########
    dates_copy = copy.deepcopy(dates)
    if len(dates) > 0:
        for each_date in dates_copy:
            #### You just create the date-time input only once and reuse the same input again and again
            date_input = keras.Input(shape=(1,), name=each_date, dtype="string")
            date_input_dict[each_date] = date_input
            all_date_inputs.append(date_input)
            all_input_names.append(each_date)
            try:
                ### for datetime strings, you need to split them into hour, day and month ######
                encoded_hour = encode_date_time_var_hourofday_categorical(date_input, each_date, train_ds)
                all_date_encoded.append(encoded_hour)
                if verbose:
                    print('    %s : after date-hour encoding shape: %s' %(each_date, encoded_hour.shape[1]))
                    if encoded_hour.shape[1] > high_cats_alert:
                        print('        Alert! excessive feature dimension created. Check if necessary to have this many.')
            except:
                print('    Error: Skipping %s since Keras Date hourofday preprocessing erroring' %each_date)
            try:
                ### for datetime strings, you need to split them into hour, day and month ######
                encoded_day = encode_date_time_var_dayofweek_categorical(date_input, each_date, train_ds)
                all_date_encoded.append(encoded_day)
                if verbose:
                    print('    %s : after date-day encoding shape: %s' %(each_date, encoded_day.shape[1]))
                    if encoded_day.shape[1] > high_cats_alert:
                        print('        Alert! excessive feature dimension created. Check if necessary to have this many.')
            except:
                print('    Error: Skipping %s since Keras Date dayofweek preprocessing erroring' %each_date)
            try:
                ### for datetime strings, you need to split them into hour, day and month ######
                encoded_month = encode_date_time_var_monthofyear_categorical(date_input, each_date, train_ds)
                all_date_encoded.append(encoded_month)
                if verbose:
                    print('    %s : after date-month encoding shape: %s' %(each_date, encoded_month.shape[1]))
                    if encoded_month.shape[1] > high_cats_alert:
                        print('        Alert! excessive feature dimension created. Check if necessary to have this many.')
            except:
                print('    Error: Skipping %s since Keras Date dayofweek preprocessing erroring' %each_date)
            #### This is where you do the category crossing of hourofday and dayofweek first 24*7 bins

            try:
                encoded_hour_day = encode_cat_feature_crosses_numeric(encoded_day, encoded_hour, train_ds,
                                            bins_num=24*7)
                all_date_encoded.append(encoded_hour_day)
                if verbose:
                    print('    %s : after date-hour-day encoding shape: %s' %(each_date, encoded_hour_day.shape[1]))
                    if encoded_hour_day.shape[1] > high_cats_alert:
                        print('        Alert! excessive feature dimension created. Check if necessary to have this many.')
            except:
                print('    Error: Skipping %s since Keras Date day-hour cross preprocessing erroring' %each_date)
            #### This is where you do the category crossing of dayofweek and monthofyear first 12*7 bins
            try:
                encoded_month_day = encode_cat_feature_crosses_numeric(encoded_month, encoded_day, train_ds,
                                            bins_num=12*7)
                all_date_encoded.append(encoded_month_day)
                if verbose:
                    print('    %s : after date-day-month encoding shape: %s' %(each_date, encoded_month_day.shape[1]))
                    if encoded_month_day.shape[1] > high_cats_alert:
                        print('        Alert! excessive feature dimension created. Check if necessary to have this many.')
            except:
                print('    Error: Skipping %s since Keras Date month-day cross preprocessing erroring' %each_date)


    #####  If boolean variables exist, you must do this here ######
    if len(bools) > 0:
        bools_copy = copy.deepcopy(bools)
        try:
            for each_bool in bools_copy:
                #### You just create the boolean input  as float since we are converting it ###
                if train_data_is_file:
                    bool_input = keras.Input(shape=(1,), name=each_bool, dtype=tf.string)
                else:
                    bool_input = keras.Input(shape=(1,), name=each_bool, dtype="float32")
                bool_input_dict[each_bool] = bool_input
                encoded = bool_input
            all_bool_encoded.append(encoded)
            all_bool_inputs.append(bool_input)
            all_input_names.append(each_bool)
            if verbose:
                print('    %s : after boolean encoding, is now float with  shape: %s' %(each_bool, encoded.shape[1]))
        except:
            print('    Error: Skipping %s since Keras Bolean preprocessing is erroring' %each_bool)

    ######  This is where we handle Boolean + Integer variables - we just combine them ##################
    int_bools_copy = copy.deepcopy(int_bools)
    if len(int_bools_copy) > 0:
        for each_int in int_bools_copy:
            try:
                int_input = keras.Input(shape=(1,), name=each_int, dtype="int32")
                cat_input_dict[each_int] = int_input
                vocab = max_tokens_zip[each_int]
                layer = tf.keras.layers.experimental.preprocessing.IntegerLookup(vocabulary=vocab,
                            mask_token=None, num_oov_indices=1, output_mode="int")
                # Convert the string input values into a one hot encoding.
                encoded = layer(int_input)
                all_int_inputs.append(int_input)
                all_int_bool_encoded.append(encoded)
                all_input_names.append(each_int)
                if verbose:
                    print('    %s number of categories = %d and after integer encoding shape: %s' %(each_int,
                                            len(max_tokens_zip[each_int]), encoded.shape[1]))
            except:
                print('    Error: Skipping %s since Keras Boolean Integer preprocessing erroring' %each_int)

    ######  This is where we handle high cardinality >50 categories integers ##################
    ints_copy = copy.deepcopy(ints)
    if len(ints_copy) > 0:
        for each_int in ints_copy:
            try:
                ### for integers that are very high cardinality, you can cut them down by half for starters
                if max_tokens_zip[each_int] <= 100:
                    nums_bin = max(5, int(max_tokens_zip[each_int]/10))
                elif max_tokens_zip[each_int] > 100 and max_tokens_zip[each_int] <= 1000:
                    nums_bin = max(10, int(max_tokens_zip[each_int]/10))
                else:
                    nums_bin = max(20, int(max_tokens_zip[each_int]/40))
                int_input = keras.Input(shape=(1,), name=each_int, dtype="int32")
                if (max_tokens_zip[each_int] >= high_cats_alert):
                    encoded = encode_any_integer_to_hash_categorical(int_input, each_int,
                                                                            train_ds, nums_bin)
                    if verbose:
                        print('    %s encoded: %d categories, %d bins. After integer HASH encoding shape = %s' %(each_int,
                                                max_tokens_zip[each_int], nums_bin, encoded.shape[1]))
                else:
                    encoded = encode_categorical_and_integer_features(int_input, each_int,
                                                                            train_ds, is_string=False)
                    if verbose:
                        print('    %s encoded: %d categories. After integer encoding shape: %s' %(each_int,
                                                max_tokens_zip[each_int], encoded.shape[1]))
                all_int_inputs.append(int_input)
                all_int_encoded.append(encoded)
                all_input_names.append(each_int)
                if verbose:
                    if (encoded.shape[1] >= high_cats_alert):
                        print('        High Dims Alert! Convert %s to float??' %each_int)
            except:
                print('    Error: Skipping %s since Keras Integer preprocessing erroring' %each_int)


    ######  This is where we handle low cardinality <=50 categories integers ##################

    ints_cat_copy = copy.deepcopy(int_cats)
    if len(ints_cat_copy) > 0:
        for each_int in ints_cat_copy:
            try:
                int_input = keras.Input(shape=(1,), name=each_int, dtype="int32")
                cat_input_dict[each_int] = int_input
                vocab = max_tokens_zip[each_int]
                #encoded = encode_integer_to_categorical_feature(int_input, each_int,
                #                                                        train_ds, vocab)
                encoded = encode_categorical_and_integer_features(int_input, each_int,
                                                                        train_ds, is_string=False)
                all_int_cat_inputs.append(int_input)
                all_int_cat_encoded.append(encoded)
                all_input_names.append(each_int)
                if verbose:
                    print('    %s encoded: %d categories. After integer encoding shape: %s' %(each_int,
                                         len(vocab), encoded.shape[1]))
                    if encoded.shape[1] > high_cats_alert:
                        if verbose:
                            print('        High Dims Alert! Convert %s to float??' %each_int)
            except:
                print('    Error: Skipping %s since Keras Integer Categorical preprocessing erroring' %each_int)

    ##### All Discrete String and Categorical features are encoded as strings  ###########
    cats_copy = copy.deepcopy(cats)
    if len(cats_copy) > 0:
        for each_cat in cats_copy:
            if each_cat in lats+lons:
                continue    ### skip if these variables are already in another list
            try:
                cat_input = keras.Input(shape=(1,), name=each_cat, dtype="string")
                cat_input_dict[each_cat] = cat_input
                vocab = max_tokens_zip[each_cat]
                max_tokens = len(vocab)
                cat_encoded = encode_categorical_and_integer_features(cat_input, each_cat,
                                                                     train_ds, is_string=True)
                #cat_encoded = encode_string_categorical_feature_categorical(cat_input, each_cat,
                #                                                     train_ds, vocab)
                all_cat_inputs.append(cat_input)
                all_cat_encoded.append(cat_encoded)
                cat_encoded_dict[each_cat] = cat_encoded
                all_input_names.append(each_cat)
                if verbose:
                    print('    %s number of categories = %d: after string to categorical encoding shape: %s' %(
                                        each_cat, max_tokens, cat_encoded.shape[1]))
                    if cat_encoded.shape[1] > high_cats_alert:
                        if verbose:
                            print('        High Dims Alert! Convert %s to float??' %each_int)
            except:
                print('    Error: Skipping %s since Keras Categorical preprocessing erroring' %each_cat)

    ##### All Discrete String and Categorical features are encoded as strings  ###########
    high_cats_copy = copy.deepcopy(high_string_vars)
    if len(high_cats_copy) > 0:
        for each_cat in high_cats_copy:
            if each_cat in lats+lons:
                continue    ### skip if these variables are already in another list
            try:
                #### The next line is not a typo: this input should be left without shape. Be Careful!
                cat_input = keras.Input(shape=(1,), name=each_cat, dtype="string")
                vocabulary = max_tokens_zip[each_cat]
                encoded = encode_any_feature_to_embed_categorical(cat_input, each_cat,
                                                                     train_ds, vocabulary)
                all_cat_inputs.append(cat_input)
                all_high_cat_encoded.append(encoded)
                cat_encoded_dict[each_cat] = encoded
                all_input_names.append(each_cat)
                if verbose:
                    print('    %s : after high cardinality cat encoding shape: %s' %(each_cat, encoded.shape[1]))
                    if encoded.shape[1] > high_cats_alert:
                        print('        Alert! excessive feature dimension created. Check if necessary to have this many.')
            except:
                print('    Error: Skipping %s since Keras Discrete Strings (high cats) preprocessing erroring' %each_cat)

    ####  If the feature crosses for categorical variables are requested, then do this here ###
    
    if cat_feature_cross_flag:
        if isinstance(cat_feature_cross_flag, str):
            if cat_feature_cross_flag in ['cat','categorical']:
                cross_cats =  copy.deepcopy(cats)
            elif cat_feature_cross_flag in ['num', 'numeric']:
                cross_cats = int_cats+int_bools
            elif cat_feature_cross_flag in ['both', 'Both']:
                cross_cats = cats + int_cats + int_bools
        else:
            #### If it is true just perform string categorical crosses ###
            cross_cats =  copy.deepcopy(cats)
        if len(cross_cats) < 2:
            print('Feature crossing requested but not many cat or int-cat or int-bool variables in data')
        else:
            print('    Performing %s feature crossing using %d variables: \n    %s' %(cat_feature_cross_flag, len(cross_cats), cross_cats))
    ##### Now you perform crosses for each kind of crosses requested ####################
    if cat_feature_cross_flag and len(cross_cats) > 1:
        feat_cross_encoded = perform_new_feature_crossing(cat_input_dict, cross_cats, train_ds)
        all_feat_cross_encoded += feat_cross_encoded
    else:
        print('    no feature crossing performed')
    ##################################################################################
    # Numerical features are treated as Numericals  ### this is a direct feed to the final layer ###
    nums_copy = left_subtract(floats,lats+lons)
    num_only_encoded = []

    if len(nums_copy) > 0:
        for each_num in nums_copy:
            try:
                num_input = keras.Input(shape=(1,), name=each_num, dtype="float32")
                ### Let's assume we don't do any encoding but use Normalization ##
                feat_mean = max_tokens_zip[each_num][0]
                feat_var = max_tokens_zip[each_num][1]
                normalizer = Normalization(mean=feat_mean, variance=feat_var)
                encoded = normalizer(num_input)
                #encoded = encode_numerical_feature_numeric(num_input, each_num, train_ds)
                all_num_inputs.append(num_input)
                all_num_encoded.append(encoded)
                num_only_encoded.append(encoded)
                all_input_names.append(each_num)
                print('    %s numeric column left as is since float' %each_num)
            except:
                print('    Error: Skipping %s due to Keras float preprocessing error' %each_num)


    # Latitude and Longitude Numerical features are Binned first and then Category Encoded #######
    lat_lon_paired_dict = dict([])
    #### Just remember that dtype of Input should match the dtype of the column! #####
    # Latitude and Longitude Numerical features are Binned first and then Category Encoded #######
    lat_lists = []
    lats_copy = copy.deepcopy(lats)
    if len(lats_copy) > 0:
        for each_lat in lats_copy:
            lat_lists += list(cat_vocab_dict[each_lat]['vocab'])
        lats_copy = copy.deepcopy(lats)
        for each_lat in lats_copy:
            try:
                bins_lat = pd.qcut(lat_lists, q=find_number_bins(cat_vocab_dict[each_lat]['vocab']),
                                   duplicates='drop', retbins=True)[1]
                ##### Now we create the inputs and the encoded outputs ######
                lat_lon_input = keras.Input(shape=(1,), name=each_lat, dtype="float32")
                all_latlon_inputs.append(lat_lon_input)
                lat_lon_encoded = encode_binning_numeric_feature_categorical(lat_lon_input, each_lat, train_ds,
                                                bins_lat=bins_lat,
                                                bins_num=len(bins_lat))
                all_latlon_encoded.append(lat_lon_encoded)
                lat_lon_paired_dict[each_lat] = lat_lon_encoded
                all_input_names.append(each_lat)
                if verbose:
                    print('    %s : after latitude binning encoding shape: %s' %(each_lat, lat_lon_encoded.shape[1]))
                    if lat_lon_encoded.shape[1] > high_cats_alert:
                        print('        Alert! excessive feature dimension created. Check if necessary to have this many.')
            except:
                print('    Error: Skipping %s since Keras latitudes var preprocessing erroring' %each_lat)

    lon_lists = []
    lons_copy = copy.deepcopy(lons)
    if len(lons_copy) > 0:
        for each_lon in lons_copy:
            lon_lists += list(cat_vocab_dict[each_lon]['vocab'])
        lons_copy = copy.deepcopy(lons)
        for each_lon in lons_copy:
            try:
                bins_lon = pd.qcut(lon_lists, q=find_number_bins(cat_vocab_dict[each_lon]['vocab']),
                                   duplicates='drop', retbins=True)[1]
                ##### Now we create the inputs and the encoded outputs ######
                lat_lon_input = keras.Input(shape=(1,), name=each_lon, dtype="float32")
                all_latlon_inputs.append(lat_lon_input)
                lat_lon_encoded = encode_binning_numeric_feature_categorical(lat_lon_input, each_lon, train_ds,
                                                bins_lat=bins_lon,
                                                bins_num=len(bins_lon))
                all_latlon_encoded.append(lat_lon_encoded)
                lat_lon_paired_dict[each_lon] = lat_lon_encoded
                all_input_names.append(each_lon)
                if verbose:
                    print('    %s : after longitude binning encoding shape: %s' %(each_lon, lat_lon_encoded.shape[1]))
                    if lat_lon_encoded.shape[1] > high_cats_alert:
                        print('        Alert! excessive feature dimension created. Check if necessary to have this many.')
            except:
                print('    Error: Skipping %s since Keras longitudes var preprocessing erroring' %each_lon)

    #### this is where you match the pairs of latitudes and longitudes to create an embedding layer
    if len(matched_lat_lons) > 0:
        matched_lat_lons_copy = copy.deepcopy(matched_lat_lons)
        for (lat_in_pair, lon_in_pair) in matched_lat_lons_copy:
            try:
                encoded_pair = encode_feature_crosses_lat_lon_numeric(lat_lon_paired_dict[lat_in_pair],
                                                              lat_lon_paired_dict[lon_in_pair],
                                                       dataset=train_ds, bins_lat=bins_lat)
                lat_lon_paired_encoded.append(encoded_pair)
                if verbose:
                    print('    %s + %s : after matched lat-lon crosses encoding shape: %s' %(lat_in_pair, lon_in_pair, encoded_pair.shape[1]))
                    if encoded_pair.shape[1] > high_cats_alert:
                        print('        Alert! excessive feature dimension created. Check if necessary to have this many.')
            except:
                print('    Error: Skipping (%s, %s) since Keras lat-lon paired preprocessing erroring' %(lat_in_pair, lon_in_pair))

    #####  SEQUENCE OF THESE INPUTS AND OUTPUTS MUST MATCH ABOVE - we gather all outputs above into a single list
    all_inputs = all_bool_inputs + all_date_inputs + all_int_inputs + all_int_cat_inputs + all_cat_inputs + all_num_inputs + all_latlon_inputs
    all_encoded = all_bool_encoded+all_high_cat_encoded+all_date_encoded+all_int_bool_encoded+all_int_encoded+all_int_cat_encoded+all_cat_encoded+all_feat_cross_encoded+all_num_encoded+all_latlon_encoded+lat_lon_paired_encoded
    ###### This is new way of processing categorical and numeric variables #########
    all_low_cat_encoded = all_int_bool_encoded+all_cat_encoded+all_int_encoded+all_int_cat_encoded  ## these are integer outputs ####
    all_numeric_encoded = all_num_encoded+all_bool_encoded+all_date_encoded+all_high_cat_encoded+all_feat_cross_encoded+all_latlon_encoded+lat_lon_paired_encoded## these are all float ##
    
    ###### This is new way of processing categorical and numeric variables #########
    #all_low_cat_encoded = all_int_bool_encoded + all_cat_encoded + all_int_cat_encoded 
    #all_numeric_encoded = all_num_encoded + all_latlon_encoded + lat_lon_paired_encoded + all_feat_cross_encoded

    
    ###### This is where we determine the size of different layers #########
    data_size = model_options['DS_LEN']
    if len(all_numeric_encoded) == 0:
        meta_numeric_len = 1
    elif len(all_numeric_encoded) == 1:
        meta_numeric_len = all_numeric_encoded[0].shape[1]
    else:
        meta_numeric_len = layers.concatenate(all_numeric_encoded).shape[1]
    data_dim = int(data_size*meta_numeric_len)
    if data_dim <= 1e6:
        dense_layer1 = max(16,int(data_dim/30000))
        dense_layer2 = max(8,int(dense_layer1*0.5))
        dense_layer3 = max(4,int(dense_layer2*0.5))
    elif data_dim > 1e6 and data_dim <= 1e8:
        dense_layer1 = max(24,int(data_dim/500000))
        dense_layer2 = max(12,int(dense_layer1*0.5))
        dense_layer3 = max(6,int(dense_layer2*0.5))
    elif data_dim > 1e8 or keras_model_type.lower() in ['big_deep', 'big deep']:
        dense_layer1 = 300
        dense_layer2 = 200
        dense_layer3 = 100
    dense_layer1 = min(300,dense_layer1)
    dense_layer2 = min(200,dense_layer2)
    dense_layer3 = min(100,dense_layer3)
    ############   D E E P   and   W I D E   M O D E L S   P R E P R O C E S S I N G ########
    ####    P R E P R O C E S S I N G   F O R  A L L   O T H E R   M O D E L S ########
    ####Concatenate all features( Numerical input )
    skip_meta_categ1 = False
    #concat_kernel_initializer = "glorot_uniform"
    concat_kernel_initializer = "he_normal"
    concat_activation = 'relu'
    concat_layer_neurons = dense_layer1
    
    ####Concatenate all categorical features( Categorical input ) #########
    if len(all_low_cat_encoded) == 0:
        skip_meta_categ1 = True
        meta_categ1 = None
    elif len(all_low_cat_encoded) == 1:
        meta_input_categ1 = all_low_cat_encoded[0]
        meta_categ1 = layers.Dense(concat_layer_neurons, kernel_initializer=concat_kernel_initializer)(meta_input_categ1)
    else:
        int_list = [x for x in all_low_cat_encoded if x.dtype in [np.int8, np.int16, np.int32, np.int64]]
        float_list = [ x for x in all_low_cat_encoded if x.dtype in [np.float32, np.float64]]
        if len(float_list) == len(all_low_cat_encoded):
            ### All of them are floats ###
            all_high_cat_encoded += float_list
        else:
            meta_input_categ1 = layers.concatenate(int_list)
            all_high_cat_encoded += float_list
            #WIDE - This Dense layer connects to input layer - Categorical Data
            meta_categ1 = layers.Dense(concat_layer_neurons, kernel_initializer=concat_kernel_initializer)(meta_input_categ1)

    skip_meta_categ2 = False
    if len(all_high_cat_encoded) == 0:
        skip_meta_categ2 = True
        meta_categ2 = None
    elif len(all_high_cat_encoded) == 1:
        meta_input_categ2 = all_high_cat_encoded[0]
        meta_categ2 = layers.Dense(concat_layer_neurons, kernel_initializer=concat_kernel_initializer)(meta_input_categ2)
        meta_categ2 = layers.BatchNormalization()(meta_categ2)
    else:
        meta_input_categ2 = layers.concatenate(all_high_cat_encoded)
        meta_categ2 = layers.Dense(concat_layer_neurons, kernel_initializer=concat_kernel_initializer)(meta_input_categ2)
        meta_categ2 = layers.BatchNormalization()(meta_categ2)

    skip_meta_numeric = False
    if len(all_numeric_encoded) == 0:
        skip_meta_numeric = True
        meta_numeric = None
    elif len(all_numeric_encoded) == 1:
        meta_input_numeric = all_numeric_encoded[0]
        meta_numeric = layers.Dense(concat_layer_neurons, kernel_initializer=concat_kernel_initializer)(meta_input_numeric)
        meta_numeric = layers.BatchNormalization()(meta_numeric)
    else:
        #### You must concatenate these encoded outputs before sending them out!
        #DEEP - This Dense layer connects to input layer - Numeric Data
        meta_input_numeric = layers.concatenate(all_numeric_encoded)
        meta_numeric = layers.Dense(concat_layer_neurons, kernel_initializer=concat_kernel_initializer)(meta_input_numeric)
        meta_numeric = layers.BatchNormalization()(meta_numeric)

    
    ####Concatenate both Wide and Deep layers
    #### in the end, you copy it into another variable called all_features so that you can easily remember name
    all_encoded_dict = list(zip([skip_meta_categ1, skip_meta_categ2, skip_meta_numeric],
                                  ['meta_categ1', 'meta_categ2', 'meta_numeric']))
    ######   This is how you concatenate the various layers ###############################
    concat_layers = []
    try:
        for (each_skip, each_encoded) in all_encoded_dict:
            # The order in which we feed the inputs is as follows: nlps + dates + ints + cats + floats + lat-lons
            if each_skip:
                ### This means that you must skip adding this layer ######
                continue
            else:
                #### This means you must add that layer ##########
                concat_layers.append(eval(each_encoded))
    except:
        print('    Error: preprocessing layers for %s models is erroring' %keras_model_type)

    if len(concat_layers) == 0:
        print('There are no cat, integer or float variables in this data set. Hence continuing...')
        all_features = []
    elif len(concat_layers) == 1:
        all_features = concat_layers[0]
    else:
        all_features = layers.concatenate(concat_layers)

    print('Time taken for preprocessing (in seconds) = %d' %(time.time()-start_time))
    return all_features, all_inputs, all_input_names

#############################################################################################
def find_number_bins(series):
    """
    Input can be a numpy array or pandas series. Otherwise it will blow up. Be careful!
    Returns the recommended number of bins for any Series in pandas
    Input must be a float or integer column. Don't send in alphabetical series!
    """
    try:
        num_of_quantiles = int(np.log2(series.nunique())+1)
    except:
        num_of_quantiles = max(2, int(np.log2(len(series)/5)))
    return num_of_quantiles
#############################################################################################
#####   Thanks to Francois Chollet for his excellent tutorial on Keras Preprocessing functions!
#####    https://keras.io/examples/structured_data/structured_data_classification_from_scratch/
#####  Some of the functions below are derived from the tutorial. I have added many more here.
############################################################################################
def encode_numerical_feature_numeric(feature_input, name, dataset):
    """
    Inputs:
    ----------
    feature_input: must be a keras.Input variable, so make sure you create a variable first for the
             column in your dataset that want to transform. Please make sure it has a
             shape of (None, 1).
    name: this is the name of the column in your dataset that you want to transform
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!

    Outputs:
    -----------
    encoded_feature: a keras.Tensor. You can use this tensor in keras models for training.
               The Tensor has a shape of (None, 1) - None indicates that it has not been
    """
    # Create a Normalization layer for our feature
    normalizer = Normalization()

    # Prepare a Dataset that only yields our feature
    feature_ds = dataset.map(lambda x, y: x[name])
    feature_ds = feature_ds.map(lambda x: tf.expand_dims(x, -1))

    # Learn the statistics of the data
    normalizer.adapt(feature_ds)

    # Normalize the input feature
    encoded_feature = normalizer(feature_input)
    return encoded_feature

###########################################################################################
def encode_binning_numeric_feature_categorical(feature, name, dataset, bins_lat, bins_num=10):
    """
    Inputs:
    ----------
    feature: must be a keras.Input variable, so make sure you create a variable first for the
             column in your dataset that want to transform. Please make sure it has a
             shape of (None, 1).
    name: this is the name of the column in your dataset that you want to transform
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!

    Outputs:
    -----------
    encoded_feature: a keras.Tensor. You can use this tensor in keras models for training.
               The Tensor has a shape of (None, 1) - None indicates that it has not been
    CategoryEncoding output dtype is float32 even if output is binary or count.
    """
    # Create a StringLookup layer which will turn strings into integer indices
    index = Discretization(bin_boundaries = bins_lat)

    # Prepare a Dataset that only yields our feature
    feature_ds = dataset.map(lambda x, y: x[name])
    feature_ds = feature_ds.map(lambda x: tf.expand_dims(x, -1))

    # If the bin_boundaries are given, then no adapt is needed since it is already known.
    #index.adapt(feature_ds)

    # Turn the string input into integer indices
    encoded_feature = index(feature)

    # Create a CategoryEncoding for our integer indices
    lat_bins = 21
    #encoder = CategoryEncoding(num_tokens=bins_num+1, output_mode="binary")
    encoder = CategoryEncoding(num_tokens=lat_bins, output_mode="binary")

    # Prepare a dataset of indices
    feature_ds = feature_ds.map(index)

    # Learn the space of possible indices
    encoder.adapt(feature_ds)

    # Apply one-hot encoding to our indices
    encoded_feature = encoder(encoded_feature)
    return encoded_feature

###########################################################################################
def encode_categorical_and_integer_features(feature, name, dataset, is_string):
    lookup_class = StringLookup if is_string else IntegerLookup
    # Create a lookup layer which will turn strings into integer indices
    lookup = lookup_class(output_mode="binary")

    # Prepare a Dataset that only yields our feature
    feature_ds = dataset.map(lambda x, y: x[name])
    feature_ds = feature_ds.map(lambda x: tf.expand_dims(x, -1))

    # Learn the set of possible string values and assign them a fixed integer index
    lookup.adapt(feature_ds)

    # Turn the string input into integer indices
    encoded_feature = lookup(feature)
    return encoded_feature
##############################################################################
def encode_string_categorical_feature_categorical(feature_input, name, dataset, vocab):
    """
    Inputs:
    ----------
    feature_input: must be a keras.Input variable, so make sure you create a variable first for the
             column in your dataset that want to transform. Please make sure it has a
             shape of (None, 1).
    name: this is the name of the column in your dataset that you want to transform
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!

    Outputs:
    -----------
    encoded_feature: a keras.Tensor. You can use this tensor in keras models for training.
               The Tensor has a shape of (None, 1) - None indicates that it is not batched.
    When the output_mode = "binary" or "count", the output is in float otherwise it is integer.
    """
    extra_oov = 3
    # Create a StringLookup layer which will turn strings into integer indices
    index = StringLookup(max_tokens=None, num_oov_indices=extra_oov,
                        vocabulary=vocab, output_mode="count")

    # Prepare a Dataset that only yields our feature
    feature_ds = dataset.map(lambda x, y: x[name])
    feature_ds = feature_ds.map(lambda x: tf.expand_dims(x, -1))

    ### turn it into an index first ###
    encoded_feature = index(feature_input)

    # Create a CategoryEncoding for our integer indices
    lat_bins = 10
    encoder = CategoryEncoding(num_tokens=lat_bins, output_mode="binary")

    # Prepare a dataset of indices
    feature_ds = feature_ds.map(index)

    # Learn the set of possible string values and assign them a fixed integer index
    encoder.adapt(feature_ds)

    # Turn the string input into integer indices
    encoded_feature = encoder(encoded_feature)

    return encoded_feature

###########################################################################################
def encode_integer_to_categorical_feature(feature_input, name, dataset, vocab):
    """
    Inputs:
    ----------
    feature_input: must be a keras.Input variable, so make sure you create a variable first for the
             column in your dataset that want to transform. Please make sure it has a
             shape of (None, 1).
    name: this is the name of the column in your dataset that you want to transform
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!

    Outputs:
    -----------
    encoded_feature: a keras.Tensor. You can use this tensor in keras models for training.
               The Tensor has a shape of (None, 1) - None indicates that it has not been
    When the output_mode = "binary" or "count", the output is in float otherwise it is integer.
    """
    extra_oov = 3
    # Create a StringLookup layer which will turn strings into integer indices
    ### For now we will leave the max_values as None which means there is no limit.
    index = IntegerLookup(vocabulary=vocab, mask_token=None,
                        num_oov_indices=extra_oov, oov_token=-9999,
                        output_mode='int')

    # Prepare a Dataset that only yields our feature
    feature_ds = dataset.map(lambda x, y: x[name])
    feature_ds = feature_ds.map(lambda x: tf.expand_dims(x, -1))

    # Learn the set of possible string values and assign them a fixed integer index
    index.adapt(feature_ds)

    # Turn the string input into integer indices
    encoded_feature = index(feature_input)

    return encoded_feature

###########################################################################################
def encode_cat_feature_crosses_numeric(encoded_input1, encoded_input2, dataset, bins_num=64):
    """
    This function does feature crosses of two categorical features sent in as encoded inputs.
    DO NOT SEND in RAW KERAS.INPUTs = they won't work here. This function takes those that are encoded.
    It then creates a feature cross, hashes the resulting categories and then category encodes them.
    The resulting output can be directly used an encoded variable for building pipelines.

    Inputs:
    ----------
    encoded_input1: This must be an encoded input - create a Keras.input variable first.
             Then do a StringLookup column on it and then a CategoryEncoding of it. Now you
             can feed that encoded variable into this as the first input.
    encoded_input1: This must be an encoded input - Similar to above: create a Keras.input variable first.
             Then do a StringLookup column on it and then a CategoryEncoding of it. Now you
             can feed that encoded variable into this as the second input.
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!
    bins_num: this is the number of bins you want to use in the hashing of the column
            Typically this can be 64. But you can make it smaller or larger.


    Outputs:
    -----------
    cat_cross_cat1_cat2: a keras.Tensor. You can use this tensor in keras models for training.
               The Tensor has a shape of (None, 1) -  None indicates it is batched.
    CategoryEncoding output dtype is float32 even if output is binary or count.
    """
    ###########   Categorical cross of two categorical features is done here    #########
    cross_cat1_cat2 = tf.keras.layers.experimental.preprocessing.CategoryCrossing()(
                                                [encoded_input1, encoded_input2])
    hash_cross_cat1_cat2 = tf.keras.layers.experimental.preprocessing.Hashing(num_bins=bins_num)(
                                                cross_cat1_cat2)
    #cat_cross_cat1_cat2 = tf.keras.layers.experimental.preprocessing.CategoryEncoding(
    #                                    num_tokens = bins_num)(hash_cross_cat1_cat2)

    # Cross to embedding
    cat_cross_cat1_cat2 = tf.keras.layers.Embedding(
                        (bins_num + 1) , 10) (hash_cross_cat1_cat2)
    cat_cross_cat1_cat2 = tf.reduce_sum(cat_cross_cat1_cat2, axis=-2)

    return cat_cross_cat1_cat2
###########################################################################################
def encode_cat_feature_crosses(encoded_input1, encoded_input2, dataset, bins_num=64):
    """
    This function does feature crosses of two categorical features sent in as encoded inputs.
    DO NOT SEND in RAW KERAS.INPUTs = they won't work here. This function takes those that are encoded.
    It then creates a feature cross, hashes the resulting categories and then category encodes them.
    The resulting output can be directly used an encoded variable for building pipelines.

    Inputs:
    ----------
    encoded_input1: This must be an encoded input - create a Keras.input variable first.
             Then do a StringLookup column on it and then a CategoryEncoding of it. Now you
             can feed that encoded variable into this as the first input.
    encoded_input1: This must be an encoded input - Similar to above: create a Keras.input variable first.
             Then do a StringLookup column on it and then a CategoryEncoding of it. Now you
             can feed that encoded variable into this as the second input.
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!
    bins_num: this is the number of bins you want to use in the hashing of the column
            Typically this can be 64. But you can make it smaller or larger.


    Outputs:
    -----------
    cat_cross_cat1_cat2: a keras.Tensor. You can use this tensor in keras models for training.
               The Tensor has a shape of (None, 1) -  None indicates it is batched.
    CategoryEncoding output dtype is float32 even if output is binary or count.
    """
    ###########   Categorical cross of two categorical features is done here    #########
    cross_cat1_cat2 = tf.keras.layers.experimental.preprocessing.CategoryCrossing()(
                                                [encoded_input1, encoded_input2])
    hash_cross_cat1_cat2 = tf.keras.layers.experimental.preprocessing.Hashing(num_bins=bins_num)(
                                                cross_cat1_cat2)
    cat_cross_cat1_cat2 = tf.keras.layers.experimental.preprocessing.CategoryEncoding(
                                        num_tokens = bins_num)(hash_cross_cat1_cat2)

    return cat_cross_cat1_cat2
###########################################################################################
def encode_feature_crosses_lat_lon_numeric(cat_pickup_lat, cat_pickup_lon, dataset, bins_lat):
    """
    This function does feature crosses of a paired latitude and logitude sent in as encoded inputs.
    DO NOT SEND in RAW KERAS.INPUTs = they won't work here. This function takes those that are encoded.
    It then creates a feature cross, hashes the resulting categories and then category encodes them.
    The resulting output can be directly used an encoded variable for building pipelines.

    Inputs:
    ----------
    cat_pickup_lat: This must be an encoded input - create a Keras.input variable first.
             Then do a Discretization column on it and then a CategoryEncoding of it. Now you
             can feed that encoded variable into this as the first input.
    cat_pickup_lon: This must be an encoded input - Similar to above: create a Keras.input variable first.
             Then do a Discretization column on it and then a CategoryEncoding of it. Now you
             can feed that encoded variable into this as the second input.
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!
    bins_lat: this is a pandas qcut bins - DO NOT SEND IN A NUMBER. It will fail!
            Typically you do this after binning the Latitude or Longitude after pd.qcut and set ret_bins=True.


    Outputs:
    -----------
    embed_cross_pick_lon_lat: a keras.Tensor. You can use this tensor in keras models for training.
               The Tensor has a shape of (None, embedding_dim) - None indicates it is batched.
    hashing creates integer outputs. So don't concatenate it with float outputs.
    """
    num_bins_lat = len(bins_lat)
    if num_bins_lat <= 100:
        nums_bin = max(5, int(num_bins_lat/10))
    elif num_bins_lat > 100 and num_bins_lat <= 1000:
        nums_bin = max(10, int(num_bins_lat/10))
    else:
        nums_bin = max(20, int(nums_bin/40))
    ### nums_bin = (len(bins_lat) + 1) ** 2 ## this was the old one
    ###########   Categorical cross of two categorical features is done here    #########
    cross_pick_lon_lat = tf.keras.layers.experimental.preprocessing.CategoryCrossing()(
                            [cat_pickup_lat, cat_pickup_lon])
    hash_cross_pick_lon_lat = tf.keras.layers.experimental.preprocessing.Hashing(
                            num_bins=nums_bin)(cross_pick_lon_lat)

    # Cross to embedding
    embed_cross_pick_lon_lat = tf.keras.layers.Embedding(
                        nums_bin, 10) (hash_cross_pick_lon_lat)
    embed_cross_pick_lon_lat = tf.reduce_sum(embed_cross_pick_lon_lat, axis=-2)

    return embed_cross_pick_lon_lat
################################################################################
def encode_any_integer_to_hash_categorical(feature_input, name, dataset, bins_num=30):
    """
    Inputs:
    ----------
    feature_input: must be a keras.Input variable, so make sure you create a variable first for the
             column in your dataset that want to transform. Please make sure it has a
             shape of (None, 1).
    name: this is the name of the column in your dataset that you want to transform
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!
    bins_num: this is the number of bins you want the hashing layer to split the data into

    Outputs:
    -----------
    encoded_feature: a keras.Tensor. You can use this tensor in keras models for training.
            The Tensor has a shape of (None, bins_num) - None indicates data has been batched
    hashing always gives you int64 encodings. So make sure you concatenate it with other integer encoders.
    """
    # Use the Hashing layer to hash the values to the range [0, 30]
    hasher = Hashing(num_bins=bins_num, salt=1337)

    # Prepare a Dataset that only yields our feature
    feature_ds = dataset.map(lambda x, y: x[name])
    feature_ds = feature_ds.map(lambda x: tf.expand_dims(x, -1))

    # Learn the set of possible string values and assign them a fixed integer index
    hasher.adapt(feature_ds)

    # Turn the string input into integer indices
    encoded_feature = hasher(feature_input)

    return encoded_feature

###########################################################################################
def encode_any_feature_to_embed_categorical(feature_input, name, dataset, vocabulary):
    """
    Inputs:
    ----------
    feature_input: must be a keras.Input variable, so make sure you create a variable first for the
             column in your dataset that want to transform. Please make sure it has a
             shape of (None, 1).
    name: this is the name of the column in your dataset that you want to transform
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!
    vocabulary: this is the number of bins you want the hashing layer to split the data into

    Outputs:
    -----------
    encoded_feature: a keras.Tensor. You can use this tensor in keras models for training.
            The Tensor has a shape of (None, bins_num) - None indicates data has been batched
    """
    extra_oov = 4
    # Use the lookup table first and then use embedding layer
    lookup = StringLookup(
        vocabulary=vocabulary,
        mask_token=None,
        num_oov_indices=extra_oov,
        max_tokens=None,
        output_mode="int"
    )
    # Convert the string input values into integer indices.
    #feature_ds = dataset.map(lambda x, y: x[name])
    #feature_ds = feature_ds.map(lambda x: tf.expand_dims(x, -1))

    # Learn the set of possible string values and assign them a fixed integer index
    #lookup.adapt(feature_ds)
    encoded_feature = lookup(feature_input)
    #embedding_dims = int(math.sqrt(len(vocabulary)))
    embedding_dims = int(max(2, math.log(len(vocabulary), 2)))
    # Create an embedding layer with the specified dimensions.
    embedding = tf.keras.layers.Embedding(
        input_dim=len(vocabulary)+extra_oov, output_dim=embedding_dims
    )
    # Convert the index values to embedding representations.
    encoded_feature = embedding(encoded_feature)
    encoded_feature = Flatten()(encoded_feature)

    return encoded_feature

###########################################################################################
def encode_date_time_var_dayofweek_categorical(feature_input, name, dataset):
    """
    This function will split the day of week from date-time column and create a new column.
    It will take a keras.Input variable as input and return a keras.layers variable as output.

    Inputs:
    ----------
    feature_input: must be a keras.Input variable, so make sure you create a variable first for the
             date-time column in your dataset that you want to transform. Please make sure it has a
             shape of (None, 1). It will split the hour of day from that column and create a new column.
    name: this is the name of the column in your dataset that you want to transform
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!

    Outputs:
    -----------
    encoded_feature: a keras.Tensor. You can use this tensor in keras models for training.
               The Tensor has a shape of (None, 1) - None indicates that it has not been
    CategoryEncoding output dtype is float32 even if output is binary or count.
    """
    index = StringLookup()

    # Prepare a Dataset that only yields our feature
    feature_ds = dataset.map(lambda x, y: dayofweek(x[name]))
    feature_ds = feature_ds.map(lambda x: tf.expand_dims(x, -1))

    # Learn the set of possible string values and assign them a fixed integer index
    index.adapt(feature_ds)

    # Turn the string input into integer indices
    encoded_feature = index(feature_input)

    # Create a CategoryEncoding for our integer indices
    encoder = CategoryEncoding(num_tokens=8, output_mode="binary")

    # Prepare a dataset of indices
    feature_ds = feature_ds.map(index)

    # Learn the space of possible indices
    encoder.adapt(feature_ds)

    # Apply one-hot encoding to our indices
    encoded_feature = encoder(encoded_feature)
    return encoded_feature

def encode_date_time_var_monthofyear_categorical(feature_input, name, dataset):
    """
    This function will split the month of year from date-time column and create a new column.
    It will take a keras.Input variable as input and return a keras.layers variable as output.

    Inputs:
    ----------
    feature_input: must be a keras.Input variable, so make sure you create a variable first for the
             date-time column in your dataset that you want to transform. Please make sure it has a
             shape of (None, 1). It will split the hour of day from that column and create a new column.
    name: this is the name of the column in your dataset that you want to transform
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!

    Outputs:
    -----------
    encoded_feature: a keras.Tensor. You can use this tensor in keras models for training.
               The Tensor has a shape of (None, 1) - None indicates that it has not been
    CategoryEncoding output dtype is float32 even if output is binary or count.
    """
    index = StringLookup()

    # Prepare a Dataset that only yields our feature
    feature_ds = dataset.map(lambda x, y: monthofyear(x[name]))
    feature_ds = feature_ds.map(lambda x: tf.expand_dims(x, -1))

    # Learn the set of possible string values and assign them a fixed integer index
    index.adapt(feature_ds)

    # Turn the string input into integer indices
    encoded_feature = index(feature_input)

    # Create a CategoryEncoding for our integer indices
    encoder = CategoryEncoding(num_tokens=13, output_mode="binary")

    # Prepare a dataset of indices
    feature_ds = feature_ds.map(index)

    # Learn the space of possible indices
    encoder.adapt(feature_ds)

    # Apply one-hot encoding to our indices
    encoded_feature = encoder(encoded_feature)
    return encoded_feature

def encode_date_time_var_hourofday_categorical(feature_input, name, dataset):
    """
    This function will split the hour of day from date-time column and create a new column.
    It will take a keras.Input variable as input and return a keras.layers variable as output.

    Inputs:
    ----------
    feature_input: must be a keras.Input variable, so make sure you create a variable first for the
             date-time column in your dataset that you want to transform. Please make sure it has a
             shape of (None, 1). It will split the hour of day from that column and create a new column.
    name: this is the name of the column in your dataset that you want to transform
    dataset: this is the variable holding the tf.data.Dataset of your data. Can be any kind of dataset.
            for example: it can be a batched or a prefetched dataset.
            Warning: You must be careful to set num_epochs when creating this dataset.
                   If num_epochs=None, this function will loop forever. If you set it to a number,
                   it will stop after that many epochs. So be careful!

    Outputs:
    -----------
    encoded_feature: a keras.Tensor. You can use this tensor in keras models for training.
               The Tensor has a shape of (None, 1) - None indicates that it has not been
    CategoryEncoding output dtype is float32 even if output is binary or count.
    """
    index = StringLookup()

    # Prepare a Dataset that only yields our feature
    feature_ds = dataset.map(lambda x, y: hourofday(x[name]))
    feature_ds = feature_ds.map(lambda x: tf.expand_dims(x, -1))

    # Learn the set of possible string values and assign them a fixed integer index
    index.adapt(feature_ds)

    # Turn the string input into integer indices
    encoded_feature = index(feature_input)

    # Create a CategoryEncoding for our integer indices
    encoder = CategoryEncoding(num_tokens=25, output_mode="binary")

    # Prepare a dataset of indices
    feature_ds = feature_ds.map(index)

    # Learn the space of possible indices
    encoder.adapt(feature_ds)

    # Apply one-hot encoding to our indices
    encoded_feature = encoder(encoded_feature)
    return encoded_feature
################################################################################
def one_hot_encode_categorical_target(features, labels, categories):
    """Returns a one-hot encoded tensor representing categorical values."""
    # The entire encoding can fit on one line:
    labels = tf.cast(tf.equal(categories, tf.reshape(labels, [-1, 1])), tf.int32)
    return (features, labels)
##############################################################################################
def convert_classifier_targets(labels):
    """
    This handy function converts target labels that are binary or multi-class (whether integer or string) into integers.
    This is similar to a label encoder in scikit-learn but works on tensorflow tf.data.Datasets.
    Just send in a tf.data.Dataset and it will split it into features and labels and then turn them into correct labels.
    It returns the converted labels and a dictionary which you can use to convert it back to original labels. Neat!
    """
    _, converted_labels = tf.unique(labels)
    return converted_labels
#########################################################################################
def compare_two_datasets_with_idcol(train_ds, validate_ds, idcol,verbose=0):
    ls_test = list(validate_ds.as_numpy_iterator())
    ls_train = list(train_ds.as_numpy_iterator())
    if verbose >= 1:
        print('    Size of dataset 1 = %d' %(len(ls_train)))
        print('    Size of dataset 2 = %d' %(len(ls_test)))
    ts_list = [ls_test[x][0][idcol] for x in range(len(ls_test)) ]
    tra_list = [ls_train[x][0][idcol] for x in range(len(ls_train)) ]
    print('Alert! %d rows in common between dataset 1 and 2' %(len(tra_list) - len(left_subtract(tra_list, ts_list))))
##########################################################################################
def process_continuous_data(data):
    # Normalize data
    max_data = tf.reduce_max(data)
    min_data = tf.reduce_max(data)
    data = (tf.cast(data, tf.float32) - min_data)/(max_data - min_data)
    return tf.reshape(data, [-1, 1])
##########################################################################################
# Process continuous features.
def preprocess(features, labels):
    for feature in floats:
        features[feature] = process_continuous_data(features[feature])
    return features, labels
##########################################################################################
######  This code is a modified version of keras.io documentation code examples ##########
######   https://keras.io/examples/structured_data/wide_deep_cross_networks/    ##########
##########################################################################################
def encode_auto_inputs(inputs, CATEGORICAL_FEATURE_NAMES, FLOATS, CATEGORICAL_FEATURES_WITH_VOCABULARY,
                         hidden_units, use_embedding=False):
    ### These are for string cat variables with small number of categories #############
    cat_encoded = []
    numeric_encoded = []
    text_encoded = []
    encoded_features = []
    #### In "auto" model, "wide" part is short. Hence we use "count" with "embedding" flag.
    for feature_name in inputs:
        vocabulary = CATEGORICAL_FEATURES_WITH_VOCABULARY[feature_name]
        extra_oov = 3
        if feature_name in CATEGORICAL_FEATURE_NAMES:
            cat_encoded.append('')
            cat_len = len(vocabulary)
            lookup = StringLookup(vocabulary=vocabulary, 
                                mask_token=None, 
                                oov_token = '~UNK~')
            if len(vocabulary) > 32:
                # Convert the string input values into integer indices.
                encoded_feature = inputs[feature_name]
                encoded_feature = lookup(encoded_feature)
                embedding_dims = int(max(2, math.log(len(vocabulary), 2)))
                # Create an embedding layer with the specified dimensions.
                embedding = Embedding(
                    input_dim=len(vocabulary)+extra_oov, output_dim=embedding_dims
                )
                # Convert the index values to embedding representations.
                encoded_feature = embedding(encoded_feature)
                cat_encoded[-1] = Flatten()(encoded_feature)
            else:
                encoded_feature = inputs[feature_name]
                encoded_feature = lookup(encoded_feature)
                cat_encoded[-1] = CategoryEncoding(num_tokens = cat_len + 1)(encoded_feature)
        elif feature_name in FLOATS:
            ### you just ignore the floats in cross models ####
            numeric_encoded.append('')
            feat_mean = CATEGORICAL_FEATURES_WITH_VOCABULARY[feature_name][0]
            feat_var = CATEGORICAL_FEATURES_WITH_VOCABULARY[feature_name][1]
            normalizer = Normalization(mean=feat_mean, variance=feat_var)
            numeric_encoded[-1] = normalizer(inputs[feature_name])
        else:
            cat_encoded.append('')
            if len(vocabulary) > 100:
                print('    ALERT! Excessive dimensions in %s. Should integer %s be a float variable?' %(
                                    len(vocabulary), feature_name))
                use_embedding = True
            lookup = IntegerLookup(
                vocabulary=vocabulary,
                mask_token=None,
                num_oov_indices=extra_oov,
                max_tokens=None,
                oov_token=-9999,
                output_mode="count" if use_embedding else "binary",
            )
            # Use the numerical features as-is.
            encoded_feature = inputs[feature_name]
            cat_encoded[-1] = lookup(encoded_feature)
    ##### This is where are float encoded features are combined ###
    ####Concatenate all features( Numerical input )
    encoded_features = layers.concatenate(cat_encoded+numeric_encoded)
    return encoded_features
##################################################################################################
import math
def encode_fast_inputs(inputs, CATEGORICAL_FEATURE_NAMES, FLOATS, CATEGORICAL_FEATURES_WITH_VOCABULARY,
                         use_embedding=False):
    encoded_features = []
    for feature_name in inputs:
        vocabulary = CATEGORICAL_FEATURES_WITH_VOCABULARY[feature_name]
        extra_oov = 3
        if feature_name in CATEGORICAL_FEATURE_NAMES:
            # Create a lookup to convert string values to an integer indices.
            # Since we are not using a mask token but expecting some out of vocabulary
            # (oov) tokens, we set mask_token to None and  num_oov_indices to extra_oov.
            if len(vocabulary) > 32:
                use_embedding = True
            lookup = StringLookup(
                vocabulary=vocabulary,
                mask_token=None,
                num_oov_indices=extra_oov,
                max_tokens=None,
                output_mode="int" if use_embedding else "binary",
            )
            if use_embedding:
                # Convert the string input values into integer indices.
                encoded_feature = inputs[feature_name]
                encoded_feature = lookup(encoded_feature)
                #embedding_dims = int(math.sqrt(len(vocabulary)))
                embedding_dims = int(max(2, math.log(len(vocabulary), 2)))
                # Create an embedding layer with the specified dimensions.
                embedding = layers.Embedding(
                    input_dim=len(vocabulary)+extra_oov, output_dim=embedding_dims
                )
                # Convert the index values to embedding representations.
                encoded_feature = embedding(encoded_feature)
                encoded_feature = Flatten()(encoded_feature)
            else:
                # Convert the string input values into a one hot encoding.
                encoded_feature = lookup(inputs[feature_name])
        elif feature_name in FLOATS:
            ### you just ignore the floats in cross models ####
            feat_mean = CATEGORICAL_FEATURES_WITH_VOCABULARY[feature_name][0]
            feat_var = CATEGORICAL_FEATURES_WITH_VOCABULARY[feature_name][1]
            normalizer = Normalization(mean=feat_mean, variance=feat_var)
            encoded_feature = normalizer(inputs[feature_name])
        else:
            if len(vocabulary) > 100:
                print('    ALERT! Excessive feature dimension in %s. Should %s be a float variable?' %(
                                    len(vocabulary), feature_name))
                use_embedding = True
            lookup = IntegerLookup(
                vocabulary=vocabulary,
                mask_token=None,
                num_oov_indices=extra_oov,
                max_tokens=None,
                oov_token=-9999,
                output_mode="count" if use_embedding else "binary",
            )
            # Use the numerical features as-is.
            encoded_feature = inputs[feature_name]
            encoded_feature = lookup(encoded_feature)
        encoded_features.append(encoded_feature)
    ##### This is where are float encoded features are combined ###
    all_features = layers.concatenate(encoded_features)
    return all_features
##########################################################################################
######  This code is a modified version of keras.io documentation code examples ##########
######   https://keras.io/examples/structured_data/wide_deep_cross_networks/    ##########
##########################################################################################
def closest(lst, K):
    """
    Find a number in list lst that is closest to the value K.
    """
    return lst[min(range(len(lst)), key = lambda i: abs(lst[i]-K))]
##############################################################################################
def create_nlp_inputs(nlp_vars):
    inputs = {}
    for feature_name in nlp_vars:
            inputs[feature_name] = layers.Input(
                name=feature_name, shape=(1,), dtype=tf.string
            )
    return inputs
#################################################################################
def encode_nlp_inputs(inputs, CATEGORICAL_FEATURES_WITH_VOCABULARY):
    layerlist = []
    list_embedding_sizes = [8, 16, 24, 32, 48, 64, 96, 128, 256]
    for feature_name in inputs:
        vocabulary = CATEGORICAL_FEATURES_WITH_VOCABULARY[feature_name]['vocab']
        extra_oov = 50
        #vocab_size = int(math.sqrt(len(vocabulary)))
        #best_embedding_size = closest(list_embedding_sizes, vocab_size//4000)
        best_embedding_size = int(max(2, math.log(len(vocabulary), 2)))
        
        lookup = StringLookup(
            vocabulary=vocabulary,
            mask_token=None,
            num_oov_indices=extra_oov,
            max_tokens=None,
            output_mode="int",
        )
        # Convert the string input values into integer indices.
        encoded_feature = inputs[feature_name]
        encoded_feature = lookup(encoded_feature)
        #embedding_dims = int(math.sqrt(len(vocabulary)))
        # Create an embedding layer with the specified dimensions.
        embedding = layers.Embedding(
            input_dim=len(vocabulary)+extra_oov, output_dim=best_embedding_size
        )
        # Convert the index values to embedding representations.
        encoded_feature = embedding(encoded_feature)
        encoded_feature = Flatten()(encoded_feature)
        layerlist.append(encoded_feature)
    ##### This is where are float encoded features are combined ###
    all_features = layers.concatenate(layerlist)
    #all_features = layers.Concatenate(axis = -1)(layerlist)
    #all_features = Flatten()(all_features)
    return all_features
#################################################################################
def create_fast_inputs(FEATURE_NAMES, NUMERIC_FEATURE_NAMES, FLOATS):
    inputs = {}
    for feature_name in FEATURE_NAMES:
        if feature_name in FLOATS:
            inputs[feature_name] = layers.Input(
                name=feature_name, shape=(1,), dtype=tf.float32
            )
        elif feature_name in NUMERIC_FEATURE_NAMES:
            inputs[feature_name] = layers.Input(
                name=feature_name, shape=(1,), dtype=tf.int32
            )
        else:
            inputs[feature_name] = layers.Input(
                name=feature_name, shape=(1,), dtype=tf.string
            )
    return inputs
#################################################################################
def create_all_inputs(FEATURE_NAMES, NUMERIC_FEATURE_NAMES, FLOATS):
    inputs = {}
    for feature_name in FEATURE_NAMES:
        if feature_name in FLOATS:
            inputs[feature_name] = layers.Input(
                name=feature_name, shape=(1,), dtype=tf.float32
            )
        elif feature_name in NUMERIC_FEATURE_NAMES:
            inputs[feature_name] = layers.Input(
                name=feature_name, shape=(1,), dtype=tf.float32
            )
        else:
            inputs[feature_name] = layers.Input(
                name=feature_name, shape=(1,), dtype=tf.string
            )
    return inputs
########################################################################################
def encode_num_inputs(inputs, CATEGORICAL_FEATURE_NAMES, FLOATS, CATEGORICAL_FEATURES_WITH_VOCABULARY,
                         use_embedding=False):
    encoded_features = []
    for feature_name in inputs:
        if feature_name in FLOATS:
            # Convert the string input values into a one hot encoding.
            encoded_feature = inputs[feature_name]
            encoded_features.append(encoded_feature)
    ##### This is where are float encoded features are combined ###
    all_features = layers.concatenate(encoded_features)
    return all_features
####################################################################################################
def encode_all_inputs(inputs, CATEGORICAL_FEATURE_NAMES, FLOATS, CATEGORICAL_FEATURES_WITH_VOCABULARY,
                         use_embedding=False):
    #### This is a new version intended to reduce dimensions #################
    encoded_features = []
    for feature_name in inputs:
        vocabulary = CATEGORICAL_FEATURES_WITH_VOCABULARY[feature_name]
        extra_oov = 3
        if feature_name in CATEGORICAL_FEATURE_NAMES:
            # Create a lookup to convert string values to an integer indices.
            # Since we are not using a mask token but expecting some out of vocabulary
            # (oov) tokens, we set mask_token to None and  num_oov_indices to extra_oov.
            if len(vocabulary) > 32:
                use_embedding = True
            lookup = StringLookup(
                vocabulary=vocabulary,
                mask_token=None,
                num_oov_indices=extra_oov,
                max_tokens=None,
                output_mode="int" if use_embedding else "binary",
            )
            if use_embedding:
                # Convert the string input values into integer indices.
                encoded_feature = inputs[feature_name]
                encoded_feature = lookup(encoded_feature)
                embedding_dims = int(max(2, math.log(len(vocabulary), 2)))
                # Create an embedding layer with the specified dimensions.
                embedding = layers.Embedding(
                    input_dim=len(vocabulary)+extra_oov, output_dim=embedding_dims
                )
                # Convert the index values to embedding representations.
                encoded_feature = embedding(encoded_feature)
                encoded_feature = Flatten()(encoded_feature)
            else:
                # Convert the string input values into a one hot encoding.
                encoded_feature = lookup(inputs[feature_name])
        else:
            ##### This is for both integer and floats = they are considered same ###
            ### you just ignore the floats in cross models ####
            feat_mean = CATEGORICAL_FEATURES_WITH_VOCABULARY[feature_name][0]
            feat_var = CATEGORICAL_FEATURES_WITH_VOCABULARY[feature_name][1]
            normalizer = Normalization(mean=feat_mean, variance=feat_var)
            encoded_feature = normalizer(inputs[feature_name])
            #encoded_feature = inputs[feature_name]
        encoded_features.append(encoded_feature)
    ###################
    int_list = [x for x in encoded_features if x.dtype in [np.int8, np.int16, np.int32, np.int64]]
    float_list = [ x for x in encoded_features if x.dtype in [np.float32, np.float64]]
    if len(int_list) > 0:
        all_int_features = layers.concatenate(int_list)
        meta_int1 = layers.Dense(32)(all_int_features)
    if len(float_list) > 0:
        all_float_features = layers.concatenate(float_list)
        meta_float1 = layers.Dense(32)(all_float_features)
    #### You can add a Dense layer if needed here ###########
    if len(int_list) > 0:
        if len(float_list) > 0:
            all_features = layers.concatenate([meta_int1, meta_float1])
        else:
            all_features = layers.concatenate([meta_int1])
    else:
        all_features = layers.concatenate([meta_float1])
    ##### This is where are float encoded features are combined ###
    return all_features
################################################################################
from itertools import combinations
def perform_new_feature_crossing(cat_input_dict, cross_cats, dataset):
    combos = combinations(cross_cats, 2)
    combos_encoded_list = []
    for cat1, cat2 in combos:
        cat_combo_encoded = encode_cat_feature_crosses(cat_input_dict[cat1], cat_input_dict[cat2], dataset, bins_num=64)
        combos_encoded_list.append(cat_combo_encoded)
    return combos_encoded_list
##################################################################################
def perform_feature_crossing(cat_input_dict, cross_cats, cats, floats, max_tokens_zip, verbose=0):

    high_cats_alert = 50 ### set this number to alery you when a variable has high dimensions. Should it?
    hidden_units = [32, 32]  ## this is the number needed for feature crossing
    dropout_rate = 0.1 ### this is set at a low rate ###
    if len(cross_cats) > 20:
        print('    Too many categorical features (>20). hence taking first 20 features for crossing.')
        ls = cross_cats[:20]
    elif len(cross_cats) != len(cat_input_dict):
        ls = cross_cats
    else:
        ls = list(cat_input_dict.keys())
    ###################################################################################
    keys = list(cat_input_dict.keys())
    for each in keys:
        if each not in ls:
            cat_input_dict.pop(each)
    cross_cats = copy.deepcopy(ls)
    ##################################################################################
    try:
        # This is a deep and cross network for cat and int-cat + int-bool feature crosses
        each_cat_coded = encode_all_inputs(cat_input_dict, cats, floats, max_tokens_zip,
                                            use_embedding=True)
        cross = each_cat_coded
        for _ in hidden_units:
            units = cross.shape[-1]
            x = layers.Dense(units)(cross)
            if cross.dtype == x.dtype:
                cross = each_cat_coded * x + cross
            else:
                each_cat_coded = tf.cast(each_cat_coded, tf.float32)
                cross = tf.cast(cross, tf.float32)
                cross = each_cat_coded * x + cross

        cross = layers.BatchNormalization()(cross)

        deep = each_cat_coded
        for units in hidden_units:
            deep = layers.Dense(units)(deep)
            deep = layers.BatchNormalization()(deep)
            deep = layers.ReLU()(deep)
            deep = layers.Dropout(dropout_rate)(deep)

        feat_cross_encoded = layers.concatenate([cross, deep])
        #feat_cross_encoded = cross
        if verbose:
            print('        feature crossing completed: cross encoding shape: %s' %(feat_cross_encoded.shape[1]))
            if feat_cross_encoded.shape[1] > high_cats_alert:
                print('        Alert! excessive feature dimension created. Check if necessary to have this many.')
    except:
        print('    Error: Skipping feature crossing since Keras preprocessing step is erroring')
    return feat_cross_encoded
########################################################################################
class OutputLayer(tf.keras.layers.Layer):
    def __init__(self):
        super(OutputLayer, self).__init__()

    def call(self, inputs, **kwargs):
        outputs = tf.keras.backend.cast(inputs, "string")
        return outputs
##########################################################################################
def encode_bool_inputs(inputs):
    encoded_features = []
    for feature_name in inputs:
        # Convert the string input values into a one hot encoding.
        lookup = StringLookup(
            vocabulary=['True', 'False'],
            mask_token=None,
            num_oov_indices=1,
            max_tokens=None,
            output_mode="binary"
        )
        encoded_feature = inputs[feature_name]
        encoded_feature = lookup(encoded_feature)
        encoded_features.append(encoded_feature)
    ##### This is where are float encoded features are combined ###
    if len(inputs) == 1:
        all_features = encoded_feature
    else:
        all_features = layers.concatenate(encoded_features)
    return all_features
############################################################################################



############################################################################################
#Copyright 2021 Google LLC

#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
############################################################################################
import pandas as pd
import numpy as np
pd.set_option('display.max_columns',500)
import matplotlib.pyplot as plt
import tempfile
import pdb
import copy
import warnings
warnings.filterwarnings(action='ignore')
import functools
# Make numpy values easier to read.
np.set_printoptions(precision=3, suppress=True)
############################################################################################
# TensorFlow ≥2.4 is required
import tensorflow as tf
np.random.seed(42)
tf.random.set_seed(42)
from tensorflow.keras import layers
from tensorflow import keras
from tensorflow.keras.layers.experimental.preprocessing import Normalization, StringLookup, CategoryCrossing
from tensorflow.keras.layers.experimental.preprocessing import IntegerLookup, CategoryEncoding
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization, Discretization, Hashing
from tensorflow.keras.layers import Embedding, Reshape, Dropout, Dense

from tensorflow.keras.optimizers import SGD, Adam, RMSprop
from tensorflow.keras import layers
from tensorflow.keras import optimizers
from tensorflow.keras.models import Model, load_model
from tensorflow.keras import callbacks
from tensorflow.keras import backend as K
from tensorflow.keras import utils
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.optimizers import SGD
from tensorflow.keras import regularizers
#####################################################################################
# Utils
from deep_autoviml.utilities.utilities import print_one_row_from_tf_dataset, print_one_row_from_tf_label
from deep_autoviml.utilities.utilities import print_classification_metrics, print_regression_model_stats
from deep_autoviml.utilities.utilities import plot_regression_residuals
from deep_autoviml.utilities.utilities import print_classification_model_stats, plot_history, plot_classification_results
from deep_autoviml.utilities.utilities import save_valid_predictions, print_classification_header
from deep_autoviml.utilities.utilities import get_callbacks, get_chosen_callback
from deep_autoviml.utilities.utilities import save_model_artifacts
from deep_autoviml.modeling.create_model import check_keras_options

#####################################################################################
from sklearn.metrics import roc_auc_score, mean_squared_error, mean_absolute_error
from IPython.core.display import Image, display
import pickle
#############################################################################################
##### Suppress all TF2 and TF1.x warnings ###################
tf2logger = tf.get_logger()
tf2logger.warning('Silencing TF2.x warnings')
tf2logger.root.removeHandler(tf2logger.root.handlers)
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
############################################################################################
from tensorflow.keras.layers import Reshape, MaxPooling1D, MaxPooling2D, AveragePooling2D, AveragePooling1D
from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Activation, Dense, Embedding, GlobalAveragePooling1D, GlobalMaxPooling1D, Dropout, Conv1D
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization
############################################################################################
#### probably the most handy function of all!
def left_subtract(l1,l2):
    lst = []
    for i in l1:
        if i not in l2:
            lst.append(i)
    return lst
##############################################################################################
import time
import os
from sklearn.metrics import balanced_accuracy_score, classification_report, confusion_matrix, roc_auc_score
import math
#########################################################################################
### Split raw_train_set into train and valid data sets first
### This is a better way to split a dataset into train and test ####
### It does not assume a pre-defined size for the data set.
def is_valid(x, y):
    return x % 5 == 0
def is_test(x, y):
    return x % 2 == 0
def is_train(x, y):
    return not is_test(x, y)
##################################################################################
def train_model(deep_model, full_ds, target, keras_model_type, keras_options,
                 model_options, var_df, cat_vocab_dict, project_name="", save_model_flag=True, 
                 verbose=0 ):
    """
    Given a keras model and a tf.data.dataset that is batched, this function will 
    train a keras model. It will first split the batched_data into train_ds and  
    valid_ds (80/20). Then it will select the right parameters based on model type and 
    train the model and evaluate it on valid_ds. It will return a keras model fully 
    trained on the full batched_data finally and train history.
    """
    ####  just use modeltype for printing that's all ###
    start_time = time.time()
    ### check the defaults for the following!
    save_model_path = model_options['save_model_path']
    save_weights_only = check_keras_options(keras_options, "save_weights_only", False)
    data_size = check_keras_options(keras_options, 'data_size', 10000)
    batch_size = check_keras_options(keras_options, 'batchsize', 64)
    num_classes = model_options["num_classes"]
    num_labels = model_options["num_labels"]
    modeltype = model_options["modeltype"]
    patience = check_keras_options(keras_options, "patience", 10)
    optimizer = keras_options['optimizer']
    class_weights = check_keras_options(keras_options, "class_weight", {})
    if not isinstance(model_options["label_encode_flag"], str):
        if not model_options["label_encode_flag"]:
            print('    removing class weights since label_encode_flag is set to False which means classes can be anything.')
            class_weights = {}
    print('    class_weights: %s' %class_weights)
    cols_len = len([item for sublist in list(var_df.values()) for item in sublist])
    print('    original datasize = %s, initial batchsize = %s' %(data_size, batch_size))
    NUMBER_OF_EPOCHS = check_keras_options(keras_options, "epochs", 100)
    if keras_options['lr_scheduler'] in ['expo', 'ExponentialDecay', 'exponentialdecay']:
        print('    chosen ExponentialDecay learning rate scheduler')
        expo_steps = (NUMBER_OF_EPOCHS*data_size)//batch_size
        learning_rate = keras.optimizers.schedules.ExponentialDecay(0.01, expo_steps, 0.1)
    else:
        learning_rate = check_keras_options(keras_options, "learning_rate", 5e-1)
    steps = max(10, (data_size//(2*batch_size)))
    print('    recommended steps per epoch = %d' %steps)
    onecycle_steps = math.ceil(data_size / batch_size) * NUMBER_OF_EPOCHS
    print('    recommended OneCycle steps = %d' %onecycle_steps)
    STEPS_PER_EPOCH = check_keras_options(keras_options, "steps_per_epoch", 
                        steps)
    #### These can be standard for every keras option that you use layers ######
    kernel_initializer = check_keras_options(keras_options, 'kernel_initializer', 'lecun_normal')
    activation='selu'
    print('    default initializer = %s, default activation = %s' %(kernel_initializer, activation))
    default_optimizer = keras.optimizers.SGD(learning_rate)
    use_bias = check_keras_options(keras_options, 'use_bias', True)
    val_monitor = keras_options['monitor']
    val_mode = keras_options['mode']
    patience = keras_options["patience"]

    if keras_options['lr_scheduler'] in ['',"onecycle", "onecycle2"]:
        #### you need to double the amount of patience for onecycle scheduler ##
        print('    Recommended: Increase patience for "onecycle" scheduler')
        patience = patience * 1.0
    callbacks_dict, tb_logpath = get_callbacks(val_mode, val_monitor, patience, learning_rate, 
                            save_weights_only, onecycle_steps, save_model_path)

    early_stopping = check_keras_options(keras_options, "early_stopping", False)
    if keras_options['lr_scheduler'] in ['expo', 'ExponentialDecay', 'exponentialdecay']:
        if early_stopping:
            callbacks_list = [callbacks_dict['early_stop'], callbacks_dict['print']]
        else:
            callbacks_list = [callbacks_dict['print']]
    else:
        chosen_callback = get_chosen_callback(callbacks_dict, keras_options)
        if not keras_options['lr_scheduler']:
            print('    chosen keras LR scheduler = default')
        else:
            print('    chosen keras LR scheduler = %s' %keras_options['lr_scheduler'])
        if keras_options['early_stopping']:
            callbacks_list = [chosen_callback,  callbacks_dict['tensor_board'], callbacks_dict['early_stop']]
        else:
            callbacks_list = [chosen_callback, callbacks_dict['tensor_board'], callbacks_dict['print']]

    print('    val mode = %s, val monitor = %s, patience = %s' %(val_mode, val_monitor, patience))
    print('    number of epochs = %d, steps per epoch = %d' %(NUMBER_OF_EPOCHS, STEPS_PER_EPOCH))
    ############## Split train into train and validation datasets here ###############
    ##################################################################################
    recover = lambda x,y: y
    print('\nSplitting train into 80+20 percent: train and validation data')
    valid_ds1 = full_ds.enumerate().filter(is_valid).map(recover)
    train_ds = full_ds.enumerate().filter(is_train).map(recover)
    heldout_ds1 = valid_ds1
    ##################################################################################
    valid_ds = heldout_ds1.enumerate().filter(is_test).map(recover)
    heldout_ds = heldout_ds1.enumerate().filter(is_test).map(recover)
    print('    Splitting validation 20 into 10+10 percent: valid and heldout data')
    ##################################################################################
    ###   V E R Y    I M P O R T A N T  S T E P   B E F O R E   M O D E L   F I T  ###    
    ##################################################################################
    shuffle_size = int(data_size)
    #shuffle_size = 100000
    print('    shuffle size = %d' %shuffle_size)
    train_ds = train_ds.cache().prefetch(batch_size).shuffle(shuffle_size, 
                            reshuffle_each_iteration=False, seed=42)#.repeat()
    valid_ds = valid_ds.prefetch(batch_size)#.repeat()

    print('Model training with best hyperparameters for %d epochs' %NUMBER_OF_EPOCHS)
    for each_callback in callbacks_list:
        print('    Callback added: %s' %str(each_callback).split(".")[-1])
    
    ############################    M O D E L     T R A I N I N G   ##################
    np.random.seed(42)
    tf.random.set_seed(42)
    history = deep_model.fit(train_ds, validation_data=valid_ds, class_weight=class_weights,
                    epochs=NUMBER_OF_EPOCHS, #steps_per_epoch=STEPS_PER_EPOCH, 
                    callbacks=callbacks_list, #validation_steps=STEPS_PER_EPOCH,
                   shuffle=False)

    print('    Model training completed. Following metrics available: %s' %history.history.keys())
    try:
        ##### this is where it stopped - you have toi subtract patience from it
        stopped_epoch = max(5,int(pd.DataFrame(history.history).shape[0] - patience))
    except:
        stopped_epoch = 100

    print('Time taken to train model (in mins) = %0.0f' %((time.time()-start_time)/60))

    #### train the model on full train data set now ###
    start_time = time.time()
    print('    Stopped epoch = %s' %stopped_epoch)

    ##################################################################################
    #######        S A V E the model here using save_model_name      #################
    ##################################################################################
    
    save_model_artifacts(deep_model, cat_vocab_dict, var_df, save_model_path, 
                        save_model_flag, model_options)
    print() ### just create an extra line after saving that is all 
    
    #################################################################################
    ########     P R E D I C T   O N   H E L D   O U T  D A T A   H E R E      ######
    #################################################################################
    try:
        if num_labels <= 1:
            y_test = np.concatenate(list(heldout_ds.map(lambda x,y: y).as_numpy_iterator()))
            print('    Single-Label: Heldout data shape: %s' %(y_test.shape,))
        else:
            iters = int(data_size/batch_size) + 1
            for inum, each_target in enumerate(target):
                add_ls = []
                for feats, labs in heldout_ds.take(iters): 
                    add_ls.append(list(labs[each_target].numpy()))
                flat_list = [item for sublist in add_ls for item in sublist]
                if inum == 0:
                    each_array = np.array(flat_list)
                else:
                    each_array = np.c_[each_array, np.array(flat_list)]
            y_test = copy.deepcopy(each_array)
            print('    Multi-Label: Heldout data shape: %s' %(y_test.shape,))
        scores = []
        ls = []
        if verbose >= 1:
            try:
                print_one_row_from_tf_label(heldout_ds)
            except:
                print('could not print samples from heldout ds labels')
        ###########################################################################
    except:
        print('Model erroring on heldout_ds predictions. Returning with model and artifacts dictionary.')
        return deep_model, cat_vocab_dict
        
    y_probas = deep_model.predict(heldout_ds)
    
    if isinstance(target, str):
        if modeltype != 'Regression':
            y_test_preds = y_probas.argmax(axis=1)
        else:
            if y_test.dtype == 'int':
                y_test_preds = y_probas.round().astype(int)
            else:
                y_test_preds = y_probas.ravel()
    else:
        if modeltype != 'Regression':
            #### This is for multi-label binary or multi-class problems ##
            for each_t in range(len(target)):
                if each_t == 0:
                    y_test_preds = y_probas[each_t].argmax(axis=1).astype(int)
                else:
                    y_test_preds = np.c_[y_test_preds, y_probas[each_t].argmax(axis=1).astype(int)]
        else:
            ### This is for Multi-Label Regression ###
            for each_t in range(len(target)):
                if each_t == 0:
                    y_test_preds = y_probas[each_t].mean(axis=1)
                else:
                    y_test_preds = np.c_[y_test_preds, y_probas[each_t].mean(axis=1)]
                if y_test.dtype == 'int':
                    y_test_preds = y_test_preds.round().astype(int)

    print('\nHeld out predictions shape:%s' %(y_test_preds.shape,))
    if verbose >= 1:
        if modeltype != 'Regression':
            print('    Sample predictions: %s' %y_test_preds[:10])
        else:
            if num_labels == 1:
                print('    Sample predictions: %s' %y_test_preds.ravel()[:10])
            else:
                print('    Sample predictions:\n%s' %y_test_preds[:10])

    #################################################################################
    ########     P L O T T I N G   V A L I D A T I O N   R E S U L T S         ######
    #################################################################################
    ###  Plot the epochs and loss metrics here #####################    
    try:
        #print('    Additionally, Tensorboard logs can be found here: %s' %tb_logpath)
        if modeltype == 'Regression':
            plot_history(history, val_monitor[4:], target)
        elif modeltype == 'Classification':
            plot_history(history, val_monitor[4:], target)
        else:
            plot_history(history, val_monitor[4:], target)
    except:
        print('    Plot history is erroring. Tensorboard logs can be found here: %s' %tb_logpath)

    print('\n###########################################################')
    print('         Held-out test data set Results:')
    num_labels = cat_vocab_dict['num_labels']
    num_classes = cat_vocab_dict['num_classes']
    if num_labels <= 1:
        #### This is for Single-Label Problems only ################################
        if modeltype == 'Regression':
            print_regression_model_stats(y_test, y_test_preds,target,plot_name=project_name)
            ### plot the regression results here #########
            plot_regression_residuals(y_test, y_test_preds, target, project_name, num_labels)
        else:
            print_classification_header(num_classes, num_labels, target)
            labels = cat_vocab_dict['original_classes']
            if cat_vocab_dict['target_transformed']:
                target_names = cat_vocab_dict['transformed_classes']
                target_le = cat_vocab_dict['target_le']
                y_pred = y_probas.argmax(axis=1)
                y_test_trans = target_le.inverse_transform(y_test)
                y_pred_trans = target_le.inverse_transform(y_pred)
                plot_classification_results(y_test_trans, y_pred_trans, labels, labels, target)
            else:
                y_pred = y_probas.argmax(axis=1)
                plot_classification_results(y_test, y_pred, labels, labels, target)
            print_classification_metrics(y_test, y_probas, proba_flag=True)
    else:
        if modeltype == 'Regression':
            #### This is for Multi-Label Regression ################################
            print_regression_model_stats(y_test, y_test_preds,target,plot_name=project_name)
            ### plot the regression results here #########
            plot_regression_residuals(y_test, y_test_preds, target, project_name, num_labels)
        else:
            #### This is for Multi-Label Classification ################################
            try:
                targets = cat_vocab_dict["target_variables"]
                for i, each_target in enumerate(targets):
                    print_classification_header(num_classes, num_labels, each_target)
                    labels = cat_vocab_dict[each_target+'_original_classes']
                    if cat_vocab_dict['target_transformed']:
                        ###### Use a nice classification matrix printing module here #########
                        target_names = cat_vocab_dict[each_target+'_transformed_classes']
                        target_le = cat_vocab_dict['target_le'][i]
                        y_pred = y_probas[i].argmax(axis=1)
                        y_test_trans = target_le.inverse_transform(y_test[:,i])
                        y_pred_trans = target_le.inverse_transform(y_pred)
                        labels = np.unique(y_test_trans) ### sometimes there is less classes
                        plot_classification_results(y_test_trans, y_pred_trans, labels, labels, each_target)
                    else:
                        y_pred = y_probas[i].argmax(axis=1)
                        labels = np.unique(y_test[:,i]) ### sometimes there are fewer classes ##
                        plot_classification_results(y_test[:,i], y_pred, labels, labels, each_target)
                    print_classification_metrics(y_test[:,i], y_probas[i], proba_flag=True)
                    #### This prints additional metrics #############
                    print(classification_report(y_test[:,i],y_test_preds[:,i]))
                    print(confusion_matrix(y_test[:,i], y_test_preds[:,i]))
            except:
                print_classification_metrics(y_test, y_test_preds, False)
                print(classification_report(y_test, y_test_preds ))

    ##################################################################################
    ###   V E R Y    I M P O R T A N T  S T E P   B E F O R E   M O D E L   F I T  ###    
    ##################################################################################
    print('\nTraining on full train dataset for %d epochs. This will take time...' %stopped_epoch)
    full_ds = full_ds.cache().shuffle(shuffle_size).prefetch(batch_size) #.repeat()
    #heldout_ds = heldout_ds.shuffle(shuffle_size).prefetch(batch_size)
    deep_model.fit(full_ds, epochs=stopped_epoch, #steps_per_epoch=STEPS_PER_EPOCH,
                 class_weight=class_weights, verbose=0)

    print('    completed. Time taken (in mins) = %0.0f' %((time.time()-start_time)/100))

    return deep_model, cat_vocab_dict
######################################################################################



############################################################################################
#Copyright 2021 Google LLC

#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
############################################################################################
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import tempfile
import pdb
import copy
import warnings
warnings.filterwarnings(action='ignore')
import functools
# Make numpy values easier to read.
np.set_printoptions(precision=3, suppress=True)
from collections import defaultdict
############################################################################################
# data pipelines and feature engg here
from deep_autoviml.models import basic, dnn, reg_dnn, dnn_drop, giant_deep, cnn1, cnn2
from deep_autoviml.preprocessing.preprocessing_tabular import encode_fast_inputs, create_fast_inputs
from deep_autoviml.preprocessing.preprocessing_tabular import encode_all_inputs, create_all_inputs
from deep_autoviml.preprocessing.preprocessing_tabular import encode_num_inputs, encode_auto_inputs

from deep_autoviml.preprocessing.preprocessing_tabular import encode_nlp_inputs, create_nlp_inputs
from deep_autoviml.preprocessing.preprocessing_nlp import aggregate_nlp_dictionaries

# Utils
from deep_autoviml.utilities.utilities import check_if_GPU_exists
from deep_autoviml.utilities.utilities import get_model_defaults, get_compiled_model, get_uncompiled_model
from deep_autoviml.utilities.utilities import check_model_options, check_keras_options
from deep_autoviml.utilities.utilities import add_outputs_to_model_body
from deep_autoviml.utilities.utilities import get_hidden_layers, add_outputs_to_auto_model_body

############################################################################################
# TensorFlow ≥2.4 is required
import tensorflow as tf
np.random.seed(42)
tf.random.set_seed(42)
from tensorflow.keras import layers
from tensorflow import keras
from tensorflow.keras.layers.experimental.preprocessing import Normalization, StringLookup
from tensorflow.keras.layers.experimental.preprocessing import IntegerLookup, CategoryEncoding
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization

from tensorflow.keras.optimizers import SGD, Adam, RMSprop
from tensorflow.keras import layers
from tensorflow.keras import optimizers
from tensorflow.keras.models import Model, load_model
from tensorflow.keras import callbacks
from tensorflow.keras import backend as K
from tensorflow.keras import utils
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.optimizers import SGD
from tensorflow.keras import regularizers

from tensorflow.keras import layers
#######################################################################################
from sklearn.metrics import roc_auc_score, mean_squared_error, mean_absolute_error
from IPython.core.display import Image, display
import pickle

##### Suppress all TF2 and TF1.x warnings ###################
tf2logger = tf.get_logger()
tf2logger.warning('Silencing TF2.x warnings')
tf2logger.root.removeHandler(tf2logger.root.handlers)
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
############################################################################################
from tensorflow.keras.layers import Reshape, MaxPooling1D, MaxPooling2D, AveragePooling2D, AveragePooling1D
from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Activation, Dense, Embedding, GlobalAveragePooling1D, GlobalMaxPooling1D, Dropout, Conv1D
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization
############################################################################################
#### probably the most handy function of all!
def left_subtract(l1,l2):
    lst = []
    for i in l1:
        if i not in l2:
            lst.append(i)
    return lst
#############################################################################################
from sklearn.metrics import balanced_accuracy_score, accuracy_score
import tensorflow as tf
from sklearn.metrics import confusion_matrix
import numpy as np
from tensorflow.python.keras import backend as K
import sys
class BalancedAccuracy(tf.keras.metrics.Metric):
    """
    ##########################################################################################
    ###### Many thanks to the source below for this Balanced Accuracy Metric #################
    ###  https://github.com/saeyslab/DeepLearning_for_ImagingFlowCytometry/blob/master/model.py
    ##########################################################################################
    """
    def __init__(self, noc, name="balanced_accuracy", **kwargs):
        super(BalancedAccuracy, self).__init__(name=name, **kwargs)

        self.noc = noc
        self.confusion_matrix = self.add_weight(
            name = "confusion_matrix",
            shape = (noc, noc),
            initializer = "zeros",
            dtype = tf.int32
        )

    def reset_states(self):
        K.batch_set_value([(v, np.zeros(shape=v.get_shape())) for v in self.variables])

    def update_state(self, y_true, y_pred, sample_weight=None):
        confusion_matrix = tf.math.confusion_matrix(y_true, tf.argmax(y_pred, axis=1), num_classes=self.noc)
        return self.confusion_matrix.assign_add(confusion_matrix)

    def result(self):
        diag = tf.linalg.diag_part(self.confusion_matrix)
        rowsums = tf.math.reduce_sum(self.confusion_matrix, axis=1)
        result = tf.math.reduce_mean(diag/rowsums, axis=0)
        return result
##########################################################################################
def create_model(use_my_model, nlp_inputs, meta_inputs, meta_outputs, nlp_outputs, keras_options, var_df,
                        keras_model_type, model_options, cat_vocab_dict):
    """
    This is a handy function to create a Sequential model architecture depending on keras_model_type option given.
    It also can re-use a model_body (without input and output layers) given by the user as input for model_body.
    It returns a model_body as well as a tuple containing a number of parameters used on defining the model and training it.
    """
    data_size = model_options['DS_LEN']
    num_classes = model_options["num_classes"]
    num_labels = model_options["num_labels"]
    modeltype = model_options["modeltype"]
    targets = cat_vocab_dict['target_variables']
    patience = check_keras_options(keras_options, "patience", 10)
    cols_len = len([item for sublist in list(var_df.values()) for item in sublist])
    if not isinstance(meta_outputs, list):
        data_dim = int(data_size*meta_outputs.shape[1])
    else:
        data_dim = int(data_size*cols_len)
    #### These can be standard for every keras option that you use layers ######
    kernel_initializer = check_keras_options(keras_options, 'kernel_initializer', 'glorot_uniform')
    activation='relu'
    if nlp_inputs:
        nlp_flag = True
    else:
        nlp_flag = False

    ##############  S E T T I N G    U P  DEEP_WIDE, DEEP_CROSS, FAST MODELS    ########################
    cats = var_df['categorical_vars']  ### these are low cardinality vars - you can one-hot encode them ##
    high_string_vars = var_df['discrete_string_vars']  ## discrete_string_vars are high cardinality vars ## embed them!
    bools = var_df['bools']
    int_cats = var_df['int_cats'] + var_df['int_bools']
    ints = var_df['int_vars']
    floats = var_df['continuous_vars']
    nlps = var_df['nlp_vars']
    lats = var_df['lat_vars']
    lons = var_df['lon_vars']
    floats = left_subtract(floats, lats+lons)

    FEATURE_NAMES = bools + cats + high_string_vars + int_cats + ints + floats
    NUMERIC_FEATURE_NAMES = int_cats + ints
    FLOATS = floats + bools
    CATEGORICAL_FEATURE_NAMES = cats + high_string_vars

    vocab_dict = defaultdict(list)
    cats_copy = copy.deepcopy(CATEGORICAL_FEATURE_NAMES+NUMERIC_FEATURE_NAMES)
    if len(cats_copy) > 0:
        for each_name in cats_copy:
            vocab_dict[each_name] = cat_vocab_dict[each_name]['vocab']

    floats_copy = copy.deepcopy(FLOATS)
    if len(floats_copy) > 0:
        for each_float in floats_copy:
            vocab_dict[each_float] = cat_vocab_dict[each_float]['vocab_min_var']

    ######################   set some defaults for model parameters here ##############
    keras_options, model_options, num_predicts, output_activation = get_model_defaults(keras_options,
                                                                    model_options, targets)
    ###### This is where you compile the model after it is built ###############
    num_classes = model_options["num_classes"]
    num_labels = model_options["num_labels"]
    modeltype = model_options["modeltype"]
    val_mode = keras_options["mode"]
    val_monitor = keras_options["monitor"]
    val_loss = keras_options["loss"]
    val_metrics = keras_options["metrics"]
    learning_rate = 5e-2
    ############################################################################
    print('Creating a(n) %s Functional model...' %keras_model_type)
    try:
        print('    number of outputs = %s, output_activation = %s' %(
                            num_labels, output_activation))
        print('    loss function: %s' %str(val_loss).split(".")[-1].split(" ")[0])
    except:
        print('    loss fn = %s    number of outputs = %s, output_activation = %s' %(
                            val_loss, num_labels, output_activation))
    try:
        #optimizer = return_optimizer(keras_options['optimizer'])
        ### you should use only string names for optimizers if strategy.scope is used
        optimizer = keras_options['optimizer']
    except:
        #####   set some default optimizers here for model parameters here ##
        if not keras_options['optimizer']:
            optimizer = keras.optimizers.SGD(learning_rate)
        elif keras_options["optimizer"] in ['RMS', 'RMSprop']:
            optimizer = keras.optimizers.RMSprop(learning_rate)
        elif keras_options['optimizer'] in ['Adam', 'adam', 'ADAM', 'NADAM', 'Nadam']:
            optimizer = keras.optimizers.Adam(learning_rate)
        else:
            optimizer = keras.optimizers.Adagrad(learning_rate)
    keras_options['optimizer'] = optimizer
    print('    initial learning rate = %s' %learning_rate)
    print('    initial optimizer = %s' %str(optimizer).split(".")[-1].split(" ")[0])
    ###################################################################################
    dense_layer1, dense_layer2, dense_layer3 = get_hidden_layers(data_dim)
    print('    Recommended hidden layers (with units in each Dense Layer)  = (%d, %d, %d)\n' %(
                                dense_layer1,dense_layer2,dense_layer3))
    fast_models = ['fast']
    fast_models1 = ['deep_and_wide','deep_wide','wide_deep',
                                'wide_and_deep','deep wide', 'wide deep', 'fast1']
    fast_models2 = ['deep_and_cross', 'deep_cross', 'deep cross', 'fast2']
    nlp_models = ['bert', 'use', 'text', 'mixed_nlp']
    #### The Deep and Wide Model is a bit more complicated. So it needs some changes in inputs! ######
    prebuilt_models = ['basic', 'simple', 'default','dnn','reg_dnn', 'deep', 'big deep',
                         'dnn_drop', 'big_deep', 'giant_deep', 'giant deep',
                        'cnn1', 'cnn','cnn2']
    ######   Just do a simple check for auto models here ####################
    preds = cat_vocab_dict["predictors_in_train"]
    NON_NLP_VARS = left_subtract(preds, nlps)
    if keras_model_type.lower() in fast_models+fast_models1+prebuilt_models+fast_models2+nlp_models:
        if len(NON_NLP_VARS) == 0:
            ## there are no non-NLP vars in dataset then just use NLP outputs Only
            all_inputs = nlp_inputs
            meta_outputs = nlp_outputs
            model_body = Sequential([layers.Dense(dense_layer3, activation='relu')])
            model_body = add_outputs_to_model_body(model_body, meta_outputs)
            model_body = get_compiled_model(all_inputs, model_body, output_activation, num_predicts,
                                    modeltype, optimizer, val_loss, val_metrics, cols_len, targets)
            print('    %s model loaded and compiled successfully...' %keras_model_type)
            print(model_body.summary())
            return model_body, keras_options
        else:
            all_inputs = nlp_inputs + meta_inputs
    else:
        ### this means it's an auto model and you create one here
        print('    creating %s model body...' %keras_model_type)
        #num_layers = check_keras_options(keras_options, 'num_layers', 1)
        model_body = tf.keras.Sequential([])
        #for l_ in range(num_layers):
        #    model_body.add(layers.Dense(dense_layer1, activation='selu', kernel_initializer="lecun_normal",
        #                              activity_regularizer=tf.keras.regularizers.l2(0.01)))
        return model_body, keras_options
    ##########################   This is for non-auto models #####################################
    if isinstance(use_my_model, str) :
        if use_my_model == "":
            if keras_model_type.lower() in ['basic', 'simple', 'default','sample model']:
                ##########  Now that we have setup the layers correctly, we can build some more hidden layers
                model_body = basic.model
            elif keras_model_type.lower() in ['reg_dnn', 'deep']:
                ##########  Now that we have setup the layers correctly, we can build some more hidden layers
                model_body = reg_dnn.model
            elif keras_model_type.lower() in ['dnn', 'simple_dnn']:
                ##########  Now that we have setup the layers correctly, we can build some more hidden layers
                model_body = dnn.model
            elif keras_model_type.lower() in ['dnn_drop', 'big_deep', 'big deep']:
                ####################################################
                model_body = dnn_drop.model
            elif keras_model_type.lower() in ['giant', 'giant_deep', 'giant deep']:
                ####################################################
                model_body = giant_deep.model
            elif keras_model_type.lower() in ['cnn', 'cnn1','cnn2']:
                ##########  Now that we have setup the layers correctly, we can build some more hidden layers
                # Conv1D + global max pooling
                if keras_model_type.lower() in ['cnn', 'cnn1']:
                    model_body = cnn1.model
                else:
                    model_body = cnn2.model
            ###### You have to do this for all prebuilt models ####################
            if keras_model_type.lower() in prebuilt_models:
                print('Adding inputs and outputs to a pre-built %s model...' %keras_model_type)
                if not isinstance(meta_outputs, list):
                    model_body = add_outputs_to_model_body(model_body, meta_outputs)
                else:
                    model_body = add_outputs_to_auto_model_body(model_body, meta_outputs, nlp_flag)
                #### This final outputs is the one that is taken into final dense layer and compiled
                print('    %s model loaded successfully. Now compiling model...' %keras_model_type)
            if keras_model_type.lower() in fast_models:
                ########## This is a simple fast model #########################
                dropout_rate = 0.1
                hidden_units = [dense_layer2, dense_layer3]
                inputs = create_fast_inputs(FEATURE_NAMES, NUMERIC_FEATURE_NAMES, FLOATS)
                features = encode_fast_inputs(inputs, CATEGORICAL_FEATURE_NAMES, FLOATS, vocab_dict,
                                use_embedding=False)
                for units in hidden_units:
                    features = layers.Dense(units)(features)
                    features = layers.BatchNormalization()(features)
                    features = layers.ReLU()(features)
                    features = layers.Dropout(dropout_rate)(features)
                all_inputs = list(inputs.values()) ### convert input layers to a list
                if len(nlps) > 0:
                    print('Starting NLP string column layer preprocessing...')
                    nlp_inputs = create_nlp_inputs(nlps)
                    max_tokens_zip, seq_tokens_zip, embed_tokens_zip, vocab_train_small = aggregate_nlp_dictionaries(nlps, cat_vocab_dict, model_options)
                    nlp_outputs = encode_nlp_inputs(nlp_inputs, cat_vocab_dict)
                    ### we call nlp_outputs as embedding in this section of the program ####
                    print('    NLP Preprocessing completed.')
                    nlp_inputs = list(nlp_inputs.values())
                    all_inputs += nlp_inputs
                    model_body = layers.concatenate([features, nlp_outputs])
                    print('    %s combined non-nlp and nlp outputs successfully...' %keras_model_type)
                else:
                    model_body = features
                    print('    %s combined non-nlp outputs successfully...' %keras_model_type)
            elif keras_model_type.lower() in fast_models1:
                ###############################################################################################
                # In a Wide & Deep model, the wide part of the model is a linear model, while the deep
                # part of the model is a multi-layer feed-forward network. We use the sparse representation
                # of the input features in the wide part of the model and the dense representation of the
                # input features for the deep part of the model.
                # VERY IMPORTANT TO NOTE that every input features contributes to both parts of the model with
                # different representations.
                ###############################################################################################
                dropout_rate = 0.1
                hidden_units = [dense_layer2, dense_layer3]
                inputs = create_all_inputs(FEATURE_NAMES, NUMERIC_FEATURE_NAMES, FLOATS)
                wide = encode_all_inputs(inputs, CATEGORICAL_FEATURE_NAMES, FLOATS, vocab_dict,
                                use_embedding=False)
                wide = layers.BatchNormalization()(wide)
                deep = encode_all_inputs(inputs, CATEGORICAL_FEATURE_NAMES, FLOATS, vocab_dict,
                                use_embedding=True)
                for units in hidden_units:
                    deep = layers.Dense(units)(deep)
                    deep = layers.BatchNormalization()(deep)
                    deep = layers.ReLU()(deep)
                    deep = layers.Dropout(dropout_rate)(deep)
                #### If there are NLP vars in dataset, you must combine them ##
                all_inputs = list(inputs.values()) ### convert input layers to a list
                if len(nlps) > 0:
                    print('Starting NLP string column layer preprocessing...')
                    nlp_inputs = create_nlp_inputs(nlps)
                    max_tokens_zip, seq_tokens_zip, embed_tokens_zip, vocab_train_small = aggregate_nlp_dictionaries(nlps, cat_vocab_dict, model_options)
                    nlp_outputs = encode_nlp_inputs(nlp_inputs, cat_vocab_dict)
                    ### we call nlp_outputs as embedding in this section of the program ####
                    print('    NLP Preprocessing completed.')
                    nlp_inputs = list(nlp_inputs.values())
                    all_inputs += nlp_inputs
                    model_body = layers.concatenate([wide, deep, nlp_outputs])
                    print('    %s combined wide, deep and nlp outputs successfully...' %keras_model_type)
                else:
                    model_body = layers.concatenate([wide, deep])
                    print('    %s combined wide and deep  successfully...' %keras_model_type)
            elif keras_model_type.lower() in fast_models2:
                ###############################################################################################
                # In a Deep & Cross model, the deep part of this model is the same as the deep part
                # created in the previous model. The key idea of the cross part is to apply explicit
                # feature crossing in an efficient way, where the degree of cross features grows with layer depth.
                ###############################################################################################
                dropout_rate = 0.1
                hidden_units = [dense_layer2, dense_layer3]
                #hidden_units = [dense_layer3]
                inputs = create_all_inputs(FEATURE_NAMES, NUMERIC_FEATURE_NAMES, FLOATS)
                x0 = encode_all_inputs(inputs, CATEGORICAL_FEATURE_NAMES, FLOATS, vocab_dict,
                                use_embedding=True)
                cross = x0
                for _ in hidden_units:
                    units = cross.shape[-1]
                    x = layers.Dense(units)(cross)
                    cross = x0 * x + cross
                cross = layers.BatchNormalization()(cross)

                deep = x0
                for units in hidden_units:
                    deep = layers.Dense(units)(deep)
                    deep = layers.BatchNormalization()(deep)
                    deep = layers.ReLU()(deep)
                    deep = layers.Dropout(dropout_rate)(deep)
                #### If there are NLP vars in dataset, you must combine them ##
                all_inputs = list(inputs.values()) ### convert input layers to a list
                if len(nlps) > 0:
                    print('Starting NLP string column layer preprocessing...')
                    nlp_inputs = create_nlp_inputs(nlps)
                    max_tokens_zip, seq_tokens_zip, embed_tokens_zip, vocab_train_small = aggregate_nlp_dictionaries(nlps, cat_vocab_dict, model_options)
                    nlp_outputs = encode_nlp_inputs(nlp_inputs, cat_vocab_dict)
                    ### we call nlp_outputs as embedding in this section of the program ####
                    print('    NLP Preprocessing completed.')
                    nlp_inputs = list(nlp_inputs.values())
                    all_inputs += nlp_inputs
                    model_body = layers.concatenate([cross, deep, nlp_outputs])
                    print('    %s combined wide, deep and nlp outputs successfully...' %keras_model_type)
                else:
                    model_body = layers.concatenate([cross, deep])
                    print('    %s combined wide and deep  successfully...' %keras_model_type)
                ################################################################################
            elif keras_model_type.lower() in nlp_models:
                print('    creating %s model body...' %keras_model_type)
                num_layers = check_keras_options(keras_options, 'num_layers', 1)
                model_body = tf.keras.Sequential([])
                for l_ in range(num_layers):
                    model_body.add(layers.Dense(dense_layer1, activation='selu', kernel_initializer="lecun_normal",
                                              #activity_regularizer=tf.keras.regularizers.l2(0.01)
                                              ))
                print('Adding inputs and outputs to a pre-built %s model...' %keras_model_type)
                if not isinstance(meta_outputs, list):
                    model_body = add_outputs_to_model_body(model_body, meta_outputs)
                else:
                    model_body = add_outputs_to_auto_model_body(model_body, meta_outputs, nlp_flag)
                #### This final outputs is the one that is taken into final dense layer and compiled
        else:
            try:
                new_module = __import__(use_my_model)
                print('Using the model given as input to build model body...')
                model_body = new_module.model
                print('    Loaded model from %s file successfully...' %use_my_model)
            except:
                print('    Loading %s model is erroring, hence building a simple sequential model with one layer...' %keras_model_type)
                ########## In case none of the loading of files works, then set up a simple model!
                model_body = Sequential([layers.Dense(dense_layer1, activation='relu')])
            ############   This is what you need to add to pre-built model body shells ###
            print('Adding inputs and outputs to a pre-built %s model...' %keras_model_type)
            if not isinstance(meta_outputs, list):
                model_body = add_outputs_to_model_body(model_body, meta_outputs)
            else:
                model_body = add_outputs_to_auto_model_body(model_body, meta_outputs, nlp_flag)
            #### This final outputs is the one that is taken into final dense layer and compiled
            print('    %s model loaded successfully. Now compiling model...' %keras_model_type)
    else:
        print('    Using your custom model given as input...')
        model_body = use_my_model
        ############   This is what you need to add to pre-built model body shells ###
        print('Adding inputs and outputs to a pre-built %s model...' %keras_model_type)
        if not isinstance(meta_outputs, list):
            model_body = add_outputs_to_model_body(model_body, meta_outputs)
        else:
            model_body = add_outputs_to_auto_model_body(model_body, meta_outputs, nlp_flag)
        #### This final outputs is the one that is taken into final dense layer and compiled
        print('    %s model loaded successfully. Now compiling model...' %keras_model_type)
    #############  You need to compile the non-auto models here ###############
    
    model_body = get_compiled_model(all_inputs, model_body, output_activation, num_predicts,
                            modeltype, optimizer, val_loss, val_metrics, cols_len, targets)
    print('    %s model loaded and compiled successfully...' %keras_model_type)
    if cols_len > 100:
        print('Too many columns to show model summary. Continuing...')
    else:
        print(model_body.summary())
    return model_body, keras_options
###############################################################################
def return_optimizer(hpq_optimizer):
    """
    This returns the keras optimizer with proper inputs if you send the string.
    hpq_optimizer: input string that stands for an optimizer such as "Adam", etc.
    """
    learning_rate_set = 5e-2
    ##### These are the various optimizers we use ################################
    momentum = keras.optimizers.SGD(lr=learning_rate_set, momentum=0.9)
    nesterov = keras.optimizers.SGD(lr=learning_rate_set, momentum=0.9, nesterov=True)
    adagrad = keras.optimizers.Adagrad(lr=learning_rate_set)
    rmsprop = keras.optimizers.RMSprop(lr=learning_rate_set, rho=0.9)
    adam = keras.optimizers.Adam(lr=learning_rate_set, beta_1=0.9, beta_2=0.999)
    adamax = keras.optimizers.Adamax(lr=learning_rate_set, beta_1=0.9, beta_2=0.999)
    nadam = keras.optimizers.Nadam(lr=learning_rate_set, beta_1=0.9, beta_2=0.999)
    best_optimizer = ''
    #############################################################################
    #### This could be turned into a dictionary but for now leave is as is for readability ##
    if hpq_optimizer.lower() == 'adam':
        best_optimizer = adam
    elif hpq_optimizer.lower() == 'sgd':
        best_optimizer = momentum
    elif hpq_optimizer.lower() == 'nadam':
        best_optimizer = nadam
    elif hpq_optimizer.lower() == 'adamax':
        best_optimizer = adamax
    elif hpq_optimizer.lower() == 'adagrad':
        best_optimizer = adagrad
    elif hpq_optimizer.lower() == 'rmsprop':
        best_optimizer = rmsprop
    else:
        best_optimizer = nesterov
    return best_optimizer
##########################################################################################



#Copyright 2021 Google LLC

#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
############################################################################################
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import tempfile
import pdb
import copy
import warnings
warnings.filterwarnings(action='ignore')
import functools
from itertools import combinations
from collections import defaultdict

# Make numpy values easier to read.
np.set_printoptions(precision=3, suppress=True)
############################################################################################
# data pipelines and feature engg here
from deep_autoviml.data_load.classify_features import check_model_options
from deep_autoviml.data_load.classify_features import find_remove_duplicates

# pre-defined TF2 Keras models and your own models here
from deep_autoviml.models.tf_hub_lookup import map_hub_to_name, map_name_to_handle
from deep_autoviml.models.tf_hub_lookup import map_name_to_preprocess

# Utils

############################################################################################
# TensorFlow ≥2.4 is required
import tensorflow as tf
np.random.seed(42)
tf.random.set_seed(42)
from tensorflow.keras import layers
from tensorflow import keras
from tensorflow.keras.layers.experimental.preprocessing import Normalization, StringLookup, Hashing
from tensorflow.keras.layers.experimental.preprocessing import IntegerLookup, CategoryEncoding, CategoryCrossing
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization, Discretization
from tensorflow.keras.layers import Embedding, Flatten

from tensorflow.keras.optimizers import SGD, Adam, RMSprop
from tensorflow.keras import layers
from tensorflow.keras import optimizers
from tensorflow.keras.models import Model, load_model
from tensorflow.keras import callbacks
from tensorflow.keras import backend as K
from tensorflow.keras import utils
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.optimizers import SGD
from tensorflow.keras import regularizers
import tensorflow_hub as hub


from sklearn.metrics import roc_auc_score, mean_squared_error, mean_absolute_error
from IPython.core.display import Image, display
import pickle
#############################################################################################
##### Suppress all TF2 and TF1.x warnings ###################
tf2logger = tf.get_logger()
tf2logger.warning('Silencing TF2.x warnings')
tf2logger.root.removeHandler(tf2logger.root.handlers)
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
############################################################################################
from tensorflow.keras.layers import Reshape, MaxPooling1D, MaxPooling2D, AveragePooling2D, AveragePooling1D
from tensorflow.keras import Model, Sequential
from tensorflow.keras.layers import Activation, Dense, Embedding, GlobalAveragePooling1D, GlobalMaxPooling1D, Dropout, Conv1D
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization
############################################################################################
#### probably the most handy function of all!
def left_subtract(l1,l2):
    lst = []
    for i in l1:
        if i not in l2:
            lst.append(i)
    return lst
##############################################################################################
import os
###########################################################################################
# We remove punctuations and HTMLs from tweets. This is done in a function,
# so that it can be passed as a parameter to the TextVectorization object.
import re
import string
def custom_standardization(input_data):
    lowercase = tf.strings.lower(input_data)
    stripped_html = tf.strings.regex_replace(lowercase, "<br />", " ")
    return tf.strings.regex_replace(
        stripped_html, "[%s]" % re.escape(string.punctuation), ""
    )
##############################################################################################
def closest(lst, K):
    """
    Find a number in list lst that is closest to the value K.
    """
    return lst[min(range(len(lst)), key = lambda i: abs(lst[i]-K))]
##############################################################################################
def preprocessing_nlp(train_ds, model_options, var_df, cat_vocab_dict, keras_model_type, verbose=0):
    """
    This produces a preprocessing layer for an incoming NLP column using TextVectorization from keras.
    You need to just send in a tf.data.DataSet from the training portion of your dataset and an nlp_column name.
    It will return a full-model-ready layer that you can add to your Keras Functional model as an NLP_layer!
    max_tokens_zip is a dictionary of each NLP column name and its max_tokens as defined by train data.
    """
    nlp_inputs = []
    all_nlp_encoded = []
    all_nlp_embeddings = []
    nlp_col_names = []
    nlp_columns = var_df['nlp_vars']
    nlp_columns =  list(set(nlp_columns))
    fast_models = ['fast']

    fast_models1 = ['deep_and_wide','deep_wide','wide_deep',
                    'wide_and_deep','deep wide', 'wide deep', 'fast1',
                    'deep_and_cross', 'deep_cross', 'deep cross', 'fast2']
    
    max_tokens_zip, seq_tokens_zip, embed_tokens_zip, vocab_train_small = aggregate_nlp_dictionaries(
                            nlp_columns, cat_vocab_dict, model_options, verbose)

    if len(nlp_columns) == 1:
        nlp_column = nlp_columns[0]
    else:
        nlp_column = 'combined_nlp_text' ### this is when there are multiple nlp columns ##

    ### Find the best sizes for various dimensions here ###########
    seq_lengths = list(seq_tokens_zip.values())
    maximum_sequence_length = max(seq_lengths)
    ## ideally you should create an unduplicated list of vocabs here and find its size
    ### the vocab_train_small holds the entire vocab of train_small data set!
    max_vocab_size = len(vocab_train_small) + 10
    best_embedding_size = max(list(embed_tokens_zip.values()))
    print('Max vocab size = %s' %max_vocab_size)

    ###### Let us set up the defauls for embedding size and max tokens to process each column
    NLP_VARS  = copy.deepcopy(nlp_columns)
    max_features = max_vocab_size ## this is the size of vocab of the whole corpus
    embedding_dim = best_embedding_size ## this is the vector size
    sequence_length = maximum_sequence_length ## this is the length of each sentence consisting of words

    #### Now perform NLP preproprocessing for each nlp_column ######
    tf_hub_model = model_options["tf_hub_model"]
    tf_hub = False
    if not tf_hub_model == "":
        print('Using Tensorflow Hub model: %s given as input' %tf_hub_model)
        tf_hub = True
    ##### This is where we use different pre-trained models to create word and sentence embeddings ##
    if keras_model_type.lower() in ['bert']:
        print('Loading %s model this will take time...' %keras_model_type)
        if os.name == 'nt':
            tfhub_path = os.path.join(keras_model_type, 'tf_cache')
            os.environ['TFHUB_CACHE_DIR'] = tfhub_path
        if tf_hub:
            tfhub_handle_encoder = model_options['tf_hub_model']
            try:
                bert_model_name = map_hub_to_name[tfhub_handle_encoder]
                tfhub_handle_preprocess = map_name_to_preprocess[bert_model_name]
            except:
                bert_model_name = 'BERT_given_by_user_input'
                tfhub_handle_preprocess = "https://tfhub.dev/tensorflow/bert_en_uncased_preprocess/3"
        else:
            bert_model_name = "BERT Uncased Small"
            tfhub_handle_preprocess = "https://tfhub.dev/tensorflow/bert_en_uncased_preprocess/3"
            tfhub_handle_encoder = 'https://tfhub.dev/tensorflow/small_bert/bert_en_uncased_L-4_H-512_A-8/2'
        preprocessor = hub.KerasLayer(tfhub_handle_preprocess, name='BERT_preprocessing')
        encoder = hub.KerasLayer(tfhub_handle_encoder, trainable=True, name='BERT_encoder')
        print(f'    {bert_model_name} selected: {tfhub_handle_encoder}')
        print(f'    Preprocessor auto-selected: {tfhub_handle_preprocess}')
    elif keras_model_type.lower() in ["use"]:
        print('Loading %s model this will take time...' %keras_model_type)
        if os.name == 'nt':
            tfhub_path = os.path.join(keras_model_type, 'tf_cache')
            os.environ['TFHUB_CACHE_DIR'] = tfhub_path
        if tf_hub:
            bert_model_name = "USE given"
            tfhub_handle_encoder = model_options['tf_hub_model']
        else:
            bert_model_name = "Universal Sentence Encoder4"
            tfhub_handle_encoder = "https://tfhub.dev/google/universal-sentence-encoder/4"
        encoder = hub.KerasLayer(tfhub_handle_encoder, 
                       input_shape=[],
                       dtype=tf.string, 
                        trainable=True, name='USE4_encoder')
        print(f'    {bert_model_name} selected: {tfhub_handle_encoder}')
    elif keras_model_type.lower() in fast_models1:
        bert_model_name = "fast NNLM 50 with Normalization"
        if os.name == 'nt':
            tfhub_path = os.path.join(keras_model_type, 'tf_cache')
            os.environ['TFHUB_CACHE_DIR'] = tfhub_path
            tfhub_handle_encoder = 'https://tfhub.dev/google/nnlm-en-dim50-with-normalization/2'
        else:
            tfhub_handle_encoder = 'https://tfhub.dev/google/nnlm-en-dim50-with-normalization/2'
        hub_layer = hub.KerasLayer(tfhub_handle_encoder,
                       input_shape=[],
                       dtype=tf.string, 
                       trainable=False, name="NNLM50_encoder")
        print(f'    {bert_model_name} selected from: {tfhub_handle_encoder}')
    elif keras_model_type.lower() in ["nlp"]:
        bert_model_name = "Swivel-20"
        if os.name == 'nt':
            tfhub_path = os.path.join(keras_model_type, 'tf_cache')
            os.environ['TFHUB_CACHE_DIR'] = tfhub_path
            tfhub_handle_encoder = 'https://tfhub.dev/google/tf2-preview/gnews-swivel-20dim/1'
        else:
            tfhub_handle_encoder = 'https://tfhub.dev/google/tf2-preview/gnews-swivel-20dim/1'
        hub_layer = hub.KerasLayer(tfhub_handle_encoder,
                       input_shape=[],
                       dtype=tf.string, 
                       trainable=False, name="Swivel20_encoder")
        print(f'    {bert_model_name} selected from: {tfhub_handle_encoder}')
    elif keras_model_type.lower() in fast_models:
        #### For fast models you just use Vectorization and Embedding that's all #######
        # Use the text vectorization layer to normalize, split, and map strings to
        # integers. Note that the layer uses the custom standardization defined above.
        # Set maximum_sequence length as all samples are not of the same length.
        ### if you used custom_standardization function, you cannot load the saved model!! be careful!
        bert_model_name = 'Text Vectorization'
        vectorize_layer = TextVectorization(
            standardize='lower_and_strip_punctuation',
            max_tokens=max_features,
            output_mode="int",
            split="whitespace",
            ngrams=None,
            output_sequence_length=sequence_length,
            pad_to_max_tokens=True,
            vocabulary=vocab_train_small,
            )
        print(f'    {bert_model_name} selected along with Embedding layer')
    else:
        ####  This is for auto model option. You can ignore their models in tfhub in that case
        #### If they give the default NLP or text as input, then we would use a default model.
        bert_model_name = 'Swivel_20_model'
        #bert_model_name = "Auto NNLM 50 with Normalization"
        tfhub_handle_encoder = "https://tfhub.dev/google/tf2-preview/gnews-swivel-20dim/1"
        #if os.name == 'nt':
        #    tfhub_path = os.path.join(keras_model_type, 'tf_cache')
        #    os.environ['TFHUB_CACHE_DIR'] = tfhub_path
        #    tfhub_handle_encoder = 'https://tfhub.dev/google/nnlm-en-dim50-with-normalization/2'
        #else:
        #    tfhub_handle_encoder = 'https://tfhub.dev/google/nnlm-en-dim50-with-normalization/2'
        hub_layer = hub.KerasLayer(tfhub_handle_encoder, output_shape=[20],
                       input_shape=[],
                       dtype=tf.string, trainable=False, name="Swivel_encoder")
        #hub_layer = hub.KerasLayer(tfhub_handle_encoder,
        #               input_shape=[],
        #               dtype=tf.string, trainable=False, name="NNLM50_encoder")
        print(f'    {bert_model_name} selected from: {tfhub_handle_encoder}')
    
    #### Next, we add an NLP layer to map those vocab indices into a space of dimensionality
    #### Vocabulary size defines how many unique words you think might be in that training data
    ### Sequence length defines how we should convert each word into a sequence of integers of fixed length

    #### Now let us process all NLP columns by using embeddings from Keras ####
    ###### A string input for each string column ###############################
    ##### Now we handle multiple choices in embedding and model building ###
    try:
        if keras_model_type.lower() in ['bert']:
            nlp_input = tf.keras.layers.Input(shape=(), dtype=tf.string, name=nlp_column)
            ### You need to do some special pre-processing if it is a BERT model
            x = encoder(preprocessor(nlp_input))['pooled_output']
        elif keras_model_type.lower() in ["use"]:
            nlp_input = tf.keras.layers.Input(shape=(), dtype=tf.string, name=nlp_column)
            ### You need to do some special pre-processing if it is a BERT model
            x = encoder(nlp_input)
        elif keras_model_type.lower() in fast_models:
            nlp_input = tf.keras.Input(shape=(), dtype=tf.string, name=nlp_column)
            x = vectorize_layer(nlp_input)
            x = layers.Embedding(max_features+1, embedding_dim, input_length=sequence_length, name=nlp_column+'_embedding')(x)
            x = Flatten()(x)
        elif keras_model_type.lower() in ["nlp"]:
            ### this is for NLP models. You use Swivel to embed NLP columns fast ####
            for each_nlp in nlp_columns:
                nlp_input = tf.keras.Input(shape=(), dtype=tf.string, name=each_nlp)
                nlp_inputs.append(nlp_input)
        elif keras_model_type.lower() in fast_models1:
            ### this is for AUTO models. You use NNLM or NLNM to embed NLP columns fast ####
            nlp_input = tf.keras.Input(shape=(), dtype=tf.string, name=nlp_column)
            x = hub_layer(nlp_input)
        else:
            ### this is for AUTO models. You use Swivel to embed NLP columns fast ####
            nlp_input = tf.keras.Input(shape=(), dtype=tf.string, name=nlp_column)
            x = hub_layer(nlp_input)
        ###  Now we combine all inputs and outputs in one place here ###########
        nlp_inputs.append(nlp_input)
        all_nlp_encoded.append(x)
        nlp_col_names.append(nlp_column)
    except:
        print('    Error: Skipping %s for keras layer preprocessing...' %nlp_column)
### we gather all outputs above into a single list here called all_features!
    if len(all_nlp_encoded) == 0:
        print('There are no NLP string variables in this dataset to preprocess!')
    elif len(all_nlp_encoded) == 1:
        all_nlp_embeddings = all_nlp_encoded[0]
    else:
        all_nlp_embeddings = layers.concatenate(all_nlp_encoded)

    return nlp_inputs, all_nlp_embeddings, nlp_col_names
###############################################################################################
def one_hot_encode_categorical_target(features, labels, categories):
    """Returns a one-hot encoded tensor representing categorical values."""
    # The entire encoding can fit on one line:
    labels = tf.cast(tf.equal(categories, tf.reshape(labels, [-1, 1])), tf.int32)
    return (features, labels)
##############################################################################################
def convert_classifier_targets(labels):
    """
    This handy function converts target labels that are binary or multi-class (whether integer or string) into integers.
    This is similar to a label encoder in scikit-learn but works on tensorflow tf.data.Datasets.
    Just send in a tf.data.Dataset and it will split it into features and labels and then turn them into correct labels.
    It returns the converted labels and a dictionary which you can use to convert it back to original labels. Neat!
    """
    _, converted_labels = tf.unique(labels)
    return converted_labels
#########################################################################################
def compare_two_datasets_with_idcol(train_ds, validate_ds, idcol,verbose=0):
    ls_test = list(validate_ds.as_numpy_iterator())
    ls_train = list(train_ds.as_numpy_iterator())
    if verbose >= 1:
        print('    Size of dataset 1 = %d' %(len(ls_train)))
        print('    Size of dataset 2 = %d' %(len(ls_test)))
    ts_list = [ls_test[x][0][idcol] for x in range(len(ls_test)) ]
    tra_list = [ls_train[x][0][idcol] for x in range(len(ls_train)) ]
    print('Alert! %d rows in common between dataset 1 and 2' %(len(tra_list) - len(left_subtract(tra_list, ts_list))))
##########################################################################################
def process_continuous_data(data):
    # Normalize data
    max_data = tf.reduce_max(data)
    min_data = tf.reduce_max(data)
    data = (tf.cast(data, tf.float32) - min_data)/(max_data - min_data)
    return tf.reshape(data, [-1, 1])
##########################################################################################
# Process continuous features.
def preprocess(features, labels):
    for feature in floats:
        features[feature] = process_continuous_data(features[feature])
    return features, labels
##########################################################################################
def encode_NLP_column(train_ds, nlp_column, nlp_input, vocab_size, sequence_length):
    text_ds = train_ds.map(lambda x,y: x[nlp_column])
    vectorize_layer = TextVectorization(
        #standardize=custom_standardization,
        standardize = 'lower_and_strip_punctuation',
        max_tokens=vocab_size,
        output_mode='int',
        output_sequence_length=sequence_length)
    # Tensorflow uses the word "adapt" to mean "fit" when learning vocabulary from a data set
    # You must call adapt first on a training data set and let it learn from that data set
    vectorize_layer.adapt(text_ds)

    ###### This is where you put NLP embedding layer into your data ####
    nlp_vectorized = vectorize_layer(nlp_input)
    ### Sometimes the get_vocabulary() errors due to special chars in utf-8. Hence avoid it.
    #print(f"    {nlp_column} vocab size = {vocab_size}, sequence_length={sequence_length}")
    return nlp_vectorized
################################################################################################
def aggregate_nlp_dictionaries(nlp_columns, cat_vocab_dict, model_options, verbose=0):
    """
    This function aggregates all the dictionaries you need for nlp processing.
    Just send in a list of nlp variables and a small data sample and it will compute all
    the seq lengths, embedding_dims and vocabs for each nlp variable in the input list.
    """
    lst = [8, 16, 24, 32, 48, 64, 96, 128, 256]
    #### max_tokens_zip calculate the max number of unique words in a vocabulary ####
    max_tokens_zip = defaultdict(int)
    #### seq_tokens_zip calculates the max sequence length in a vocabulary ####
    seq_tokens_zip = defaultdict(int)
    #### embed_tokens_zip calculates the embedding dimension for each nlp_column ####
    embed_tokens_zip = defaultdict(int)
    #### This carries the 
    nlps_copy = copy.deepcopy(nlp_columns)
    seq_lengths = []
    vocab_train_small = []
    if len(nlps_copy) > 0:
        vocab_train_small = []
        for each_name in nlps_copy:
            if verbose >= 2:
                print('Creating aggregate_nlp_dictionaries for nlp column = %s' %each_name)
            max_tokens_zip[each_name] = cat_vocab_dict[each_name]['size_of_vocab']
            print('    size of vocabulary = %s' %max_tokens_zip[each_name])
            seq_tokens_zip[each_name] = cat_vocab_dict[each_name]['seq_length']
            seq_lengths.append(seq_tokens_zip[each_name])
            if verbose >= 2:
                print('    sequence length = %s' %seq_tokens_zip[each_name])
            vocab_size = cat_vocab_dict[each_name]['size_of_vocab']
            vocab_train_small += cat_vocab_dict[each_name]['vocab']
            vocab_train_small = np.unique(vocab_train_small).tolist()
            best_embedding_size = closest(lst, vocab_size//50000)
            if verbose >= 2:
                print('    recommended embedding_size = %s' %best_embedding_size)
            input_embedding_size = check_model_options(model_options, "embedding_size", best_embedding_size)
            if input_embedding_size != best_embedding_size:
                if verbose >= 2:
                    print('    input embedding size given as %d. Overriding recommended embedding_size...' %input_embedding_size)
                best_embedding_size = input_embedding_size
            embed_tokens_zip[each_name] = best_embedding_size
    return max_tokens_zip, seq_tokens_zip, embed_tokens_zip, vocab_train_small
##################################################################################################